from __future__ import annotations


class Permission:
    """Permission Object"""

    def __init__(self, data: int) -> None:
        self.perms = {
            "CREATE_INSTANT_INVITE": 0x1,
            "KICK_MEMBERS": 0x2,
            "BAN_MEMBERS": 0x4,
            "ADMINISTRATOR": 0x8,
            "MANAGE_CHANNELS": 0x10,
            "MANAGE_GUILD": 0x20,
            "ADD_REACTIONS": 0x40,
            "VIEW_AUDIT_LOG": 0x80,
            "PRIORITY_SPEAKER": 0x100,
            "STREAM": 0x200,
            "VIEW_CHANNEL": 0x400,
            "SEND_MESSAGES": 0x800,
            "SEND_TTS_MESSAGES": 0x1000,
            "MANAGE_MESSAGES": 0x2000,
            "EMBED_LINKS": 0x4000,
            "ATTACH_FILES": 0x8000,
            "READ_MESSAGE_HISTORY": 0x10000,
            "MENTION_EVERYONE": 0x20000,
            "USE_EXTERNAL_EMOJIS": 0x40000,
            "VIEW_GUILD_INSIGHTS": 0x80000,
            "CONNECT": 0x100000,
            "SPEAK": 0x200000,
            "MUTE_MEMBERS": 0x400000,
            "DEAFEN_MEMBERS": 0x800000,
            "MOVE_MEMBERS": 0x1000000,
            "USE_VAD": 0x2000000,
            "CHANGE_NICK": 0x4000000,
            "MANAGE_NICK": 0x8000000,
            "MANAGE_ROLES": 0x10000000,
            "MANAGE_WEBHOOKS": 0x20000000,
            "MANAGE_EMOJIS_AND_STICKERS": 0x40000000,
            "USE_APPLICATION_COMMANDS": 0x80000000,
            "REQUEST_TO_SPEAK": 0x100000000,
            "MANAGE_EVENTS": 0x200000000,
            "MANAGE_THREADS": 0x400000000,
            "CREATE_PUBLIC_THREADS": 0x800000000,
            "CREATE_PRIVATE_THREADS": 0x1000000000,
            "USE_EXTERNAL_STICKERS": 0x2000000000,
            "SEND_MESSAGES_IN_THREADS": 0x4000000000,
            "USE_EMBEDDED_ACTIVITIES": 0x8000000000,
            "MODERATE_MEMBERS": 0x10000000000,
        }
        self.permissions = self.calculate_permissions(data)

    def calculate_permissions(self, perm_value: int):
        return [
            perm
            for perm, value in self.perms.items()
            if (perm_value & value) == value
        ]



import pytest
import selfcord

bot = selfcord.Bot(prefixes=["!"])

@pytest.mark.asyncio
class Tests:
    async def test(self, *args):
        return "HELLO WORLD"

    def add_cmd_test(self):
        try:
            bot.add_cmd(coro=self.test, description="Sends hello!", aliases=["hello"])
        except Exception as e:
            raise RuntimeError(f"Failed to add command. Error: {e}")

        cmds = len(bot.commands)
        assert cmds > 0








from __future__ import annotations


class Webhook:
    def __init__(self, data: dict, bot, http) -> None:
        self.http = http
        self.bot = bot
        self._update(data)

    def __eq__(self, other):
        return self.id == other.id

    def _update(self, data):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """
        self.id = data.get("id")
        self.type = data.get("type")
        self.guild_id = data.get("guild_id")
        self.channel_id = data.get("channel_id")
        self.name = data.get("name")
        self.avatar = data.get("avatar")
        self.token = data.get("token")
        self.application_id = data.get("application_id")
        self.webhook_url = f"https://discord.com/api/webhooks/{self.id}/{self.token}"
        self.source_guild = data.get("source_guild")
        self.source_channel = data.get("source_channel")

    async def send(self, content: str):
        """Send a message via the webhook

        Args:
            content (str): Content of the message to send
        """
        await self.http.request(
            method="post",
            endpoint=f"/webhooks/{self.id}/{self.token}",
            json={"content": content},
        )

    async def delete(self):
        """Deletes the webhook object"""
        await self.http.request(
            method="delete", endpoint=f"/webhooks/{self.id}/{self.token}"
        )



from .channel import (Category, DMChannel, GroupChannel, Messageable,
                      TextChannel, Voiceable, VoiceChannel)
from .client import Client
from .emoji import Emoji
from .guild import Guild
from .interactions import InteractionUtil, Option, Search, SlashCommand
from .member import Member
from .message import Message
from .permission import Permission
from .role import Role
from .user import Connected_Account, Profile, User
from .webhook import Webhook



from __future__ import annotations

import asyncio
import random
from base64 import b64encode
from traceback import format_exception
from typing import TYPE_CHECKING

import aiohttp
from aiohttp import ClientSession

from ..models import Client, User
from ..utils import logging

log = logging.getLogger("HTTP")


class http:
    """Base HTTP class to aid making requests via discord api"""

    def __init__(self, debug=False) -> None:
        self.debug = debug
        xproperties = [
            "eyJvcyI6IldpbmRvd3MiLCJicm93c2VyIjoiRmlyZWZveCIsImRldmljZSI6IiIsInN5c3RlbV9sb2NhbGUiOiJmciIsImJyb3dzZXJfdXNlcl9hZ2VudCI6Ik1vemlsbGEvNS4wIChXaW5kb3dzIE5UIDEwLjA7IFdpbjY0OyB4NjQ7IHJ2OjEwMi4wKSBHZWNrby8yMDEwMDEwMSBGaXJlZm94LzEwMi4wIiwiYnJvd3Nlcl92ZXJzaW9uIjoiMTAyLjAiLCJvc192ZXJzaW9uIjoiMTAiLCJyZWZlcnJlciI6IiIsInJlZmVycmluZ19kb21haW4iOiIiLCJyZWZlcnJlcl9jdXJyZW50IjoiIiwicmVmZXJyaW5nX2RvbWFpbl9jdXJyZW50IjoiIiwicmVsZWFzZV9jaGFubmVsIjoic3RhYmxlIiwiY2xpZW50X2J1aWxkX251bWJlciI6MTU0MTg2LCJjbGllbnRfZXZlbnRfc291cmNlIjpudWxsfQ==",
            "eyJvcyI6IkxpbnV4IiwiYnJvd3NlciI6IkRpc2NvcmQgQ2xpZW50IiwicmVsZWFzZV9jaGFubmVsIjoiY2FuYXJ5IiwiY2xpZW50X3ZlcnNpb24iOiIwLjAuMTQwIiwib3NfdmVyc2lvbiI6IjUuMTkuMC0zLXJ0MTAtTUFOSkFSTyIsIm9zX2FyY2giOiJ4NjQiLCJzeXN0ZW1fbG9jYWxlIjoiZW4tR0IiLCJ3aW5kb3dfbWFuYWdlciI6IktERSx1bmtub3duIiwiZGlzdHJvIjoiXCJNYW5qYXJvIExpbnV4XCIiLCJjbGllbnRfYnVpbGRfbnVtYmVyIjoxNTQyMTYsImNsaWVudF9ldmVudF9zb3VyY2UiOm51bGx9",
            "eyJvcyI6IkxpbnV4IiwiYnJvd3NlciI6IkNocm9tZSIsImRldmljZSI6IiIsInN5c3RlbV9sb2NhbGUiOiJlbi1HQiIsImJyb3dzZXJfdXNlcl9hZ2VudCI6Ik1vemlsbGEvNS4wIChYMTE7IExpbnV4IHg4Nl82NCkgQXBwbGVXZWJLaXQvNTM3LjM2IChLSFRNTCwgbGlrZSBHZWNrbykgQ2hyb21lLzEwNi4wLjAuMCBTYWZhcmkvNTM3LjM2IiwiYnJvd3Nlcl92ZXJzaW9uIjoiMTA2LjAuMC4wIiwib3NfdmVyc2lvbiI6IiIsInJlZmVycmVyIjoiaHR0cHM6Ly93d3cucm9ibG94LmNvbS8iLCJyZWZlcnJpbmdfZG9tYWluIjoid3d3LnJvYmxveC5jb20iLCJyZWZlcnJlcl9jdXJyZW50IjoiIiwicmVmZXJyaW5nX2RvbWFpbl9jdXJyZW50IjoiIiwicmVsZWFzZV9jaGFubmVsIjoic3RhYmxlIiwiY2xpZW50X2J1aWxkX251bWJlciI6MTU0MTg2LCJjbGllbnRfZXZlbnRfc291cmNlIjpudWxsfQ==",
        ]
        self.cookies = {}
        self.token = None
        self.xproperties = random.choice(xproperties)
        self.base_url = "https://discord.com/api/v9"

    async def static_login(self, token: str):
        """Used to retrieve basic token information

        Args:
            token (str): User token

        Returns:
            Client: A Client object
        """
        await self.get_cookie()
        self.token = token
        data = await self.request("get", "/users/@me")
        self.client = Client(data)
        if self.debug:
            log.debug("Finished Static login")
            log.info(f"Gathered information on {self.client}")
        return data

    async def get_cookie(self):
        """Gather cookie for user upon client start"""
        async with aiohttp.ClientSession() as session:
            async with session.get(
                "https://discord.com",
                headers={
                    "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) discord/0.0.139 Chrome/91.0.4472.164 Electron/13.6.6 Safari/537.36"
                },
            ) as resp:
                dcf = resp.headers["set-cookie"].split("__dcfduid=")[0].split(";")[0]
                sdc = resp.headers["set-cookie"].split("__sdcfduid=")[0].split(";")[0]
                cfr = resp.headers["set-cookie"].split("__cfruid=")[0].split(";")[0]

                self.cookies["dcf"] = self.cookies.get("dcf") if dcf != "" else ""
                self.cookies["sdc"] = self.cookies.get("sdc") if sdc != "" else ""
                self.cookies["cfr"] = self.cookies.get("cfr") if cfr != "" else ""
                self.cookie = set(self.cookies)

            async with session.get(
                "https://discord.com/api/v9/experiments",
                headers={
                    "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) discord/0.0.139 Chrome/91.0.4472.164 Electron/13.6.6 Safari/537.36"
                },
            ) as resp:
                json = await resp.json()
                self.fingerprint = json["fingerprint"]
        if self.debug:
            log.debug("Gathered cookie and fingerprint")
            log.info(f"Fingerprint Gathered: {self.fingerprint}")

    def remove_dupes(self, dictionary: dict):
        return set(dictionary)

    async def request(self, method: str, endpoint: str, *args, **kwargs) -> dict:
        """Used to send requests

        Args:
            method (str): HTTP method
            endpoint (str): Discord api endpoint

        Raises:
            LoginFailure: If you suck

        Returns:
            dict: Data, json data
        """
        url = self.base_url + endpoint
        headers = {
            "cookie": f"{self.cookie}",
            "authorization": self.token,
            "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) discord/0.0.139 Chrome/91.0.4472.164 Electron/13.6.6 Safari/537.36",
            "Content-Type": "application/json",
            "X-Super-Properties": self.xproperties,
            "X-Discord-Locale": "en-GB",
            "X-context-properties": "eyJsb2NhdGlvbiI6Ikludml0ZSBCdXR0b24gRW1iZWQiLCJsb2NhdGlvbl9ndWlsZF9pZCI6bnVsbCwibG9jYXRpb25fY2hhbm5lbF9pZCI6IjEwOTkwOTMxODEyNTUxNDM1MjUiLCJsb2NhdGlvbl9jaGFubmVsX3R5cGUiOjEsImxvY2F0aW9uX21lc3NhZ2VfaWQiOiIxMTE2NTE0MDMyODk2MTgwMjU0In0=",
            "connection": "keep-alive",
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-origin",
            "origin": "https://discord.com",
            "DNT": "1",
            "x-debug-options": "logGatewayEvents,logOverlayEvents,logAnalyticsEvents,bugReporterEnabled",
            "x-fingerprint": self.fingerprint,
            "x-discord-timezone": "Europe/London",
            "TE": "trailers",
        }

        async with ClientSession(
            timeout=aiohttp.ClientTimeout(
                total=10000, connect=10000, sock_read=10000, sock_connect=10000
            ),
            connector=aiohttp.TCPConnector(
                limit=0, limit_per_host=0, ttl_dns_cache=300
            ),
            headers=headers,
        ) as session:
            request = getattr(session, method)
            while True:
                async with request(url, *args, **kwargs) as resp:
                    if self.debug:
                        log.debug(f"Sent Request Method: {method} URL: {url} Payload: {args} {kwargs}")

                    if resp.status == 429:
                        try:
                            json = await resp.json()
                            await asyncio.sleep(json["retry_after"])
                            if self.debug:
                                log.error(f"429 Ratelimited: {json}")
                            continue
                        except Exception as e:
                            error = "".join(format_exception(e, e, e.__traceback__))
                            text = await resp.text()
                            log.error(f"Error upon parsing json : {text}")
                            if self.debug:
                                log.error(f"Error upon parsing json : \n{error}")
                                log.info(
                                    f"Attempted to send request to URL: {url} PAYLOAD: {kwargs}"
                                )
                            break

                    elif resp.status == 401:
                        json = await resp.json()
                        log.error(f"{json} -- {resp.status}")
                        break

                    elif resp.status == 403:
                        json = await resp.json()
                        log.error(f"403 Unauthorized: {json}")
                        if self.debug:
                            log.info(
                                f"Attempted to send request to URL: {url} PAYLOAD: {args} {kwargs}"
                            )
                        break

                    elif resp.status == 201:
                        data = await resp.json()
                        break

                    elif resp.status == 204:
                        data = await resp.text()
                        break

                    elif resp.ok:
                        data = await resp.json()
                        break

                    else:
                        if self.debug:
                            log.error(f"Unknown Error: {resp.status}")
                            log.info(
                                f"Attempted to send request to URL: {url} PAYLOAD: {args} {kwargs}"
                            )
                        try:
                            json = await resp.json()
                            log.error(f"Error Response: {json}")
                            break
                        except Exception as e:
                            error = "".join(format_exception(e, e, e.__traceback__))
                            log.error(f"Unable to log response: \n{error}")
                            break
                        try:
                            text = await resp.text()
                            log.error(f"{text} -- {resp.status}")
                            break
                        except:
                            log.error(f"{resp.status}")
                            break
        try:
            if resp.headers["set-cookie"]:
                dcf = resp.headers["set-cookie"].split("__dcfduid=")[0].split(";")[0]
                sdc = resp.headers["set-cookie"].split("__sdcfduid=")[0].split(";")[0]
                cfr = resp.headers["set-cookie"].split("__cfruid=")[0].split(";")[0]
                bm = resp.headers["set-cookie"].split("__cf_bm=")[0].split(";")[0]

                self.cookies["dcf"] = self.cookies.get("dcf") if dcf != "" else ""
                self.cookies["sdc"] = self.cookies.get("sdc") if sdc != "" else ""
                self.cookies["cfr"] = self.cookies.get("cfr") if cfr != "" else ""
                self.cookies["bm"] = self.cookies.get("bm") if bm != "" else ""
                self.cookie = set(self.cookies)
        except:
            pass

        try:
            return data
        except:
            return None

    async def encode_image(self, url: str) -> str:
        """Turn an image url into a b64 payload

        Args:
            url (str): The URL of the image

        Returns:
            str: The b64 payload
        """
        async with ClientSession() as session:
            async with session.get(f"{url}") as resp:
                image = b64encode(await resp.read())
                newobj = str(image).split("'", 2)
        if self.debug:
            log.debug("Finished encoding image")
            log.info(f"Encoded Image: {url}")
        return f"data:image/png;base64,{newobj[1]}"



from __future__ import annotations

import asyncio
import os
import random
from traceback import format_exception
from typing import TYPE_CHECKING

import aiofiles
import aiohttp

from ..utils import logging
from .message import Message
from .user import User
from .webhook import Webhook

if TYPE_CHECKING:
    from ..api.http import http
    from ..bot import Bot
    from ..models.permission import Permission

log = logging.getLogger("Channel")


class Messageable:
    """Parent class specific for those classes that include a textchat for sending messages."""

    def __init__(self, http: http, bot: Bot) -> None:
        self.http: http = http
        self.bot: Bot = bot

    @property
    def make_nonce(self):
        """Generate pseudorandom number."""
        return str(random.randint(0, 100000000))

    async def history(self, amount: int = 100) -> list[Message] | None:
        """
        Get channel message history.

        Args:
            amount(int) : Amount of messages to gather. Default is 100.

        Returns:
            messages(list) : List of messages from the channel.
            None : If client does not have view permission for the channel or no data found
        """
        if amount >= 100:
            data = await self.http.request(
                method="get", endpoint=f"/channels/{self.id}/messages?limit=100"
            )
        else:
            data = await self.http.request(
                method="get", endpoint=f"/channels/{self.id}/messages?limit={amount + 5}"
            )

        if data is None:
            return None

        messages = [Message(msg, self.bot, self.http) for msg in data]
        while True:
            

            data = await self.http.request(
                method="get",
                endpoint=f"/channels/{self.id}/messages?limit=100&before={data[-1]['id']}",
            )

            if len(data) <= 0:
                break
            messages.extend(Message(msg, self.bot, self.http) for msg in data)
            if len(messages) >= amount:
                break

        return messages

    async def upload_image(self, paths: list[str]) -> list[dict[str, int | str]]:
        files = []
        id = 0
        for path in paths:
            async with aiofiles.open(path, "rb") as f:
                file = await f.read()
                size = len(file)
                name = os.path.basename(path)
            files.append({"file_size" : size, "filename": f"{name}", "id": id})
            id += 1
        json = await self.http.request("post", f"/channels/{self.id}/attachments", json={"files": files})
        items = []
        for key, atch in enumerate(json['attachments']):
            upload_url = atch['upload_url']
            id = atch['id']
            upload_filename = atch['upload_filename']
            async with aiohttp.ClientSession() as session:
                async with aiofiles.open(paths[key], "rb") as f:
                    file = await f.read()
                async with session.put(upload_url, data=file): 
                    pass
            items.append({"uploaded_filename": upload_filename, "filename": os.path.basename(upload_filename) , "id": id})
        return items

        

    async def purge(self, amount: int = 0) -> None:
        """
        Delete a number of messages, starting from the most recent.

        Args:
            amount(int) : Number of messages to purge/delete.

        Returns:
            No return value
        """
        messages = await self.history(amount)
        if messages is None:
            return
        msgs = [msg for msg in messages if str(msg.author.id) == str(self.bot.user.id)]
        if amount != 0:
            for i in range(0, len(msgs[:amount]), 3):
                await asyncio.gather(
                    *(
                        asyncio.create_task(message.delete())
                        for message in msgs[:amount][i : i + 3]
                    )
                )
                await asyncio.sleep(0.3)
            return


        for i in range(0, len(msgs), 3):
            await asyncio.gather(
                *(
                    asyncio.create_task(message.delete())
                    for message in msgs[i : i + 3]
                )
            )
            await asyncio.sleep(0.3)



    async def spam(self, amount: int, content: str, file_paths: list[str] = [], tts=False) -> None:
        """
        Send multiple of the same message.

        Args:
            - amount(int) : Number of spam messages to send.
            - content(str) : The message to send.
            - tts(bool) = False : Specify whether it is a TTS message.

        Returns:
            No return value.
        """
        amount: list[int] = [i + 1 for i in range(amount)]
        for i in range(0, len(amount), 3):
            await asyncio.gather(
                *(
                    asyncio.create_task(self.send(tts=tts, content=content, file_paths=file_paths))
                    for _ in amount[i : i + 3]
                )
            )
            await asyncio.sleep(0.3)

    async def send(self, content=None, file_paths: list[str] = [], tts=False) -> Message:
        """
        Send a message to the text channel.

        Args:
            - content(str) : Message content. Should be string type or similar. Discord `embed` type is not allowed.
            - tts(bool) : Specify whether message is text-to-speech or not

        Returns:
            Message object.
        """
        json = {"content": content, "nonce": self.make_nonce, "tts": tts}
        if file_paths != []:
            vals = await self.upload_image(file_paths)
            json |= {"attachments" : vals}

        if hasattr(self, "guild_id"):
            resp = await self.http.request(
                method="post",
                endpoint=f"/channels/{self.id}/messages",
                headers={
                    "origin": "https://discord.com",
                    "referer": f"https://discord.com/channels/{self.guild_id}/{self.id}",
                },
                json=json
            )
            resp.update({"guild_id": self.guild_id})

        else:
            resp = await self.http.request(
                method="post",
                endpoint=f"/channels/{self.id}/messages",
                headers={
                    "origin": "https://discord.com",
                    "referer": f"https://discord.com/channels/{self.id}",
                },
                json=json
            )

        return Message(resp, self.bot, self.http)

    async def reply(self, message: Message, content: str, file_paths: list[str] = [], tts=False) -> Message:
        """Reply to a specific message

        Args:
            message (str): Message to reply to
            content (_type_, optional): Message content to reply with. Defaults to None.
            tts (bool, optional): Specify whether message is text-to-speech or not. Defaults to False.

        Returns:
            Message object.
        """
        json = {
            "content": content,
            "tts": tts,
            "nonce": self.make_nonce,
            "message_reference": {
                "channel_id": f"{self.id}",
                "message_id": f"{message.id}",
            },
        }

        if file_paths != []:
            vals = await self.upload_image(file_paths)
            json |= {"attachments" : vals}

        if hasattr(self, "guild_id"):
            resp = await self.http.request(
                method="post",
                endpoint=f"/channels/{self.id}/messages",
                headers={
                    "origin": "https://discord.com",
                    "referer": f"https://discord.com/channels/{self.guild_id}/{self.id}",
                },
                json=json
            )
            resp.update({"guild_id": self.guild_id})
        else:
            resp = await self.http.request(
                method="post",
                endpoint=f"/channels/{self.id}/messages",
                headers={
                    "origin": "https://discord.com",
                    "referer": f"https://discord.com/channels/{self.id}",
                },
                json=json
            )
        return Message(resp, self.bot, self.http)


class Voiceable(Messageable):
    """Parent class specific for those classes that include a voice chat, or call functionality"""

    def __init__(self, http: http, bot: Bot) -> None:
        super().__init__(http, bot)
        self.http: http = http
        self.bot: Bot = bot

    async def video_call(self):
        """Initiates a video call on the specified channel"""
        if hasattr(self, "guild_id"):
            await self.bot.gateway.video_call(self.id, self.guild_id)
        else:
            await self.bot.gateway.video_call(self.id)

    async def stream_call(self):
        """Initiates a stream call on the specified channel"""
        if hasattr(self, "guild_id"):
            await self.bot.gateway.stream_call(self.id, self.guild_id)
        else:
            await self.bot.gateway.stream_call(self.id)

    async def call(self):
        """Initiates a call on the specified channel"""

        if hasattr(self, "guild_id"):
            await self.bot.gateway.call(self.id, self.guild_id)
        else:
            await self.bot.gateway.call(self.id)

    async def leave_call(self):
        """Leaves call on the specified channel"""
        await self.bot.gateway.leave_call()


class TextChannel(Messageable):
    """
    Text Channel Object
        Represents a Guild/Server channel within Discord.
        All methods are coroutines and thus need to be awaited.

    Returns:
        Name of the channel

    Functions:
        - delete()
        - edit(name[str], parent_id[int], position[int], topic[str])
        - history()
        - purge(amount[int])
        - spam(amount[int], content[str], tts[bool]=False)
        - send()
        - reply()
        - create_webhook()

    Example Object (ref: https://discord.com/developers/docs/resources/channel#channel-object-example-guild-text-channel)
        {
          "id": "41771983423143937",
          "guild_id": "41771983423143937",
          "name": "general",
          "type": 0,
          "position": 6,
          "permission_overwrites": [],
          "rate_limit_per_user": 2,
          "nsfw": true,
          "topic": "24/7 chat about how to gank Mike #2",
          "last_message_id": "155117677105512449",
          "parent_id": "399942396007890945",
          "default_auto_archive_duration": 60
        }
    """

    def __init__(self, data, bot: Bot, http: http) -> None:
        super().__init__(http, bot)
        self.permissions: list[Permission] = []
        self.webhooks: list[Webhook] = []
        self.http: http = http
        self.bot: Bot = bot
        self._update(data)

    def __str__(self) -> str:
        return f"{self.name}"

    def __eq__(self, other):
        return self.id == other.id

    def _update(self, data: dict):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """

        self.topic = data.get("topic")
        self.rate_limit_per_user = data.get("rate_limit_per_user")
        self.position = data.get("position")
        self.name = data.get("name")
        self.id = data.get("id")
        self.guild_id = data.get("guild_id")
        self.last_message_id = data.get("last_message_id")
        self.flags = data.get("flags")
        self.default_thread_rate_limit_per_user = data.get(
            "default_thread_rate_limit_per_user"
        )
        self.category_id = data.get("parent_id")

    async def delete(self):
        """
        Deletes the text channel object.

        Args:
            No arguments required

        Returns:
            No return value
        """
        await self.http.request(method="delete", endpoint=f"/channels/{self.id}")
        del self
    
    async def edit(
        self,
        name: str = None,
        parent_id: int = None,
        position: int = None,
        topic: str = None,
    ):
        """
        Edits the text channel object details. Requires the `Manage Channels` permission.
        Not all details can be modified.

        Args:
            - name(str) : Optional - Specifies a new name for the channel object. Defaults to None.
            - parent_id(int) : Optional - Specifies a new parent (category) for the text channel object.
            - position(int) : Optional - Modifies the sorting position of the text channel within the guild.
            - topic(str) : Optional - Modifies the channel topic (0-4096 characters for GUILD_FORUM channels,
                                        0-1024 characters for all others)
        Notice: Each parent category can contain up to 50 channels.

        Returns:
            e(str) : Exception parsed as string value.
        """
        payload = {}
        if name is not None:
            payload["name"] = name
        if parent_id is not None:
            payload["parent_id"] = parent_id
        if position is not None:
            payload["position"] = position
        if topic is not None and topic != "":
            payload["topic"] = topic

        try:
            await self.http.request(
                method="patch", endpoint=f"/channels/{self.id}", json=payload
            )
        except Exception as e:
            error = "".join(format_exception(e, e, e.__traceback__))
            log.error(f"Could not edit channel \n{error}")

    async def create_webhook(self, name: str = None, avatar_url: str = None) -> Webhook:
        """
        Creates a webhook in the specified channel

        Args:
            name (str, optional): Name of the webhook. Defaults to None.
            avatar_url (str, optional): Avatar of the webhook. Requires a URL. Defaults to None.

        Returns:
            webhook (Webhook): Returns the created webhook object.

        Raises:
            NameError: Name is required
        """
        fields = {}
        if name != None:
            fields["name"] = name
        else:
            log.error("Name is required")
        if avatar_url != None:
            data = await self.http.encode_image(avatar_url)
            fields["avatar"] = data
        data = await self.http.request(
            method="post", endpoint=f"/channels/{self.id}/webhooks", json=fields
        )
        webhook = Webhook(data, self.bot, http=self.http)
        self.webhooks.append(webhook)
        return webhook

    async def create_invite(self, max_age: int = 0, max_uses: int = 0) -> str:
        """
        Creates an invite in the specified channel

        Args:
            max_age (int, optional) Seconds to invite expiration. Defaults to 0 (infinite).
            max_uses (int, optional) Maximum uses to invite expiration. Defaults to 0 (infinite).

        Returns:
            invite (str): Returns the channel invite link
        """
        data = await self.http.request(
            method="post",
            endpoint=f"/channels/{self.id}/invites",
            json={
                "max_age": max_age,
                "max_uses": max_uses,
                "target_type": None,
                "temporary": False,
                "flags": 0,
            },
        )
        return "https://discord.gg/" + data["code"]


class VoiceChannel(Voiceable, Messageable):
    """Voice Channel Object"""

    def __init__(self, data: dict, bot: Bot, http: http) -> None:
        super().__init__(http, bot)
        self.permissions: list[Permission] = []
        self.webhooks: list[Webhook] = []
        self.http = http
        self.bot = bot
        self._update(data)

    def __str__(self) -> str:
        return f"{self.name}"

    def __eq__(self, other):
        return self.id == other.id

    def _update(self, data: dict):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """
        self.name = data.get("name")
        self.id = data.get("id")
        self.guild_id = data.get("guild_id")
        self.last_message_id = data.get("last_message_id")
        self.rtc_region = data.get("rtc_region")
        self.flags = data.get("flags")
        self.bitrate = data.get("bitrate")
        self.rate_limit_per_user = data.get("rate_limit_per_user")
        self.position = data.get("position")
        self.category_id = data.get("parent_id")

    async def delete(self):
        """
        Deletes the voice channel object.
        """
        await self.http.request(method="delete", endpoint=f"/channels/{self.id}")
        del self

    async def create_webhook(self, name: str = None, avatar_url: str = None) -> Webhook:
        """
        Creates a webhook in the specified channel

        Args:
            name (str, optional): Name of the webhook. Defaults to None.
            avatar_url (str, optional): Avatar of the webhook. Requires a URL. Defaults to None.

        Returns:
            webhook (Webhook): Returns the created webhook object.

        Raises:
            NameError: Name is required
        """

        fields = {}
        if name != None:
            fields["name"] = name
        else:
            log.error("Name is required...")
        if avatar_url != None:
            data = await self.http.encode_image(avatar_url)
            fields["avatar"] = data
        data = await self.http.request(
            method="post", endpoint=f"/channels/{self.id}/webhooks", json=fields
        )
        webhook = Webhook(data, self.bot, self.http)
        self.webhooks.append(webhook)
        return webhook

    async def create_invite(self, max_age: int = 0, max_uses: int = 0) -> str:
        """
        Creates an invite in the specified channel

        Args:
            max_age (int, optional) Seconds to invite expiration. Defaults to 0 (infinite).
            max_uses (int, optional) Maximum uses to invite expiration. Defaults to 0 (infinite).

        Returns:
            invite (str): Returns the channel invite link
        """
        data = await self.http.request(
            method="post",
            endpoint=f"/channels/{self.id}/invites",
            json={
                "max_age": max_age,
                "max_uses": max_uses,
                "target_type": None,
                "temporary": False,
                "flags": 0,
            },
        )
        return "https://discord.gg/" + data["code"]


class Category:
    """Category Object"""

    def __init__(self, data: dict, bot: Bot, http: http) -> None:
        self.bot: Bot = bot
        self.http: http = http
        self.permissions: list[Permission] = []
        self._update(data)

    def __str__(self) -> str:
        return f"{self.name}"

    def __eq__(self, other):
        return self.id == other.id

    def _update(self, data: dict):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """
        self.name = data.get("name")
        self.id = data.get("id")
        self.guild_id = data.get("guild_id")
        self.position = data.get("position")
        self.flags = data.get("flags")

    async def delete(self):
        """Deletes the Category object."""
        await self.http.request(method="delete", endpoint=f"/channels/{self.id}")
        del self


class DMChannel(Voiceable, Messageable):
    """DM Channel Object"""

    def __init__(self, data: dict, bot: Bot, http: http) -> None:
        super().__init__(http, bot)
        self.http: http = http
        self.bot: Bot = bot
        self._update(data)

    def __str__(self) -> str:
        return f"{self.recipient}"

    def __eq__(self, other):
        return self.id == other.id

    def _update(self, data: dict):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """
        self.recipient = User(data.get("recipients")[0], self.bot, self.http)
        self.last_message_id = data.get("last_message_id")
        self.id = data.get("id")
        self.flags = data.get("id")

    async def delete(self):
        """Deletes the DM Channel object."""
        await self.http.request(
            method="delete", endpoint=f"/channels/{self.id}?silent=false"
        )
        del self


class GroupChannel(Voiceable, Messageable):
    """Group Channel Object"""

    def __init__(self, data: dict, bot: Bot, http: http) -> None:
        super().__init__(http, bot)
        self.recipients = []
        self.http: http = http
        self.bot: Bot = bot
        self._update(data)

    def __str__(self) -> str:
        return f"{self.name}"

    def __eq__(self, other):
        return self.id == other.id

    def _update(self, data: dict):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """
        for user in data.get("recipients"):
            self.recipients.append(User(user, self.bot, self.http))
        self.name = data.get("name")
        self.owner_id = data.get("owner_id")
        self.last_message_id = data.get("last_message_id")
        self.id = data.get("id")
        self.flags = data.get("flags")
        self.icon = data.get("icon")

    async def delete(self):
        """Deletes the Group Channel Object"""
        await self.http.request(
            method="delete", endpoint=f"/channels/{self.id}?silent=true"
        )
        del self

    async def create_invite(self) -> str:
        """
        Creates an invite in the specified group channel

        Args:
            None

        Returns:
            invite (str): Returns the channel invite link
        """
        data = await self.http.request(
            method="post",
            endpoint=f"/channels/{self.id}/invites",
            json={"max_age": 86400},
        )
        return "https://discord.gg/" + data["code"]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..api.http import http
    from ..bot import Bot


class Emoji:
    """Emoji Object"""

    def __init__(self, data: dict, bot: Bot, http: http) -> None:
        self.bot: Bot = bot
        self.http: http = http
        self._update(data)

    def __str__(self) -> str:
        return f"{self.name}"

    def __eq__(self, other):
        return self.id == other.id

    def _update(self, data: dict):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """
        self.name = data.get("name")
        self.id = data.get("id")
        self.roles = data.get("roles")
        self.managed = data.get("managed")
        self.available = data.get("available")
        self.animated = data.get("animated")
        self.guild_id = data.get("guild_id")

    async def delete(self):
        """Deletes the Emoji Object"""
        await self.http.request(
            method="delete", endpoint=f"/guilds/{self.guild_id}/emojis/{self.id}"
        )
        del self



from __future__ import annotations

import asyncio
import json
import socket
import struct
import time

import websockets
from aioconsole import aprint

try:
    import nacl.secret
    import nacl.utils
    import opuslib
except:
    pass
import io
import os
from traceback import format_exception
from typing import TYPE_CHECKING, Any

import aiofiles

# import ffmpeg
from ...utils import logging

if TYPE_CHECKING:
    from ...api import http
    from ...bot import Bot


log = logging.getLogger("Voice")


class Voice:
    """Voice class used to interact with voice websockets and send data"""

    SAMPLING_RATE = 48000
    CHANNELS = 2
    FRAME_LENGTH = 20  # in milliseconds
    SAMPLE_SIZE = struct.calcsize("h") * CHANNELS
    SAMPLES_PER_FRAME = int(SAMPLING_RATE / 1000 * FRAME_LENGTH)
    FRAME_SIZE = SAMPLES_PER_FRAME * SAMPLE_SIZE

    """Op codes"""
    READY = 2
    SESSION_DESCRIPTION = 4
    HEARTBEAT_ACK = 6
    HEARTBEAT = 8

    def __init__(
        self,
        session_id: str,
        token: str,
        endpoint: str,
        server_id: str,
        bot: Bot,
        debug=False,
    ) -> None:
        self.session_id: str = session_id
        self.token: str = token
        self.endpoint: str = endpoint
        self.server_id: str = server_id
        self.bot: Bot = bot
        self.alive = False
        self.mode = "xsalsa20_poly1305"
        self.sequence = 0
        self.debug: bool = debug
        self.timestamp = 0

    async def connect(self):
        """Connect to discord voice ws"""
        self.ws = await websockets.connect(
            f"wss://{self.endpoint}", origin="https://discord.com"
        )
        self.alive = True

    async def recv_msg(self):
        """
        Receives Message from websocket, encodes as json and runs tasks
        """
        item = json.loads(await self.ws.recv())
        op = item.get("op")  # Op code
        data = item.get("d")  # Data
        if op == self.READY:
            await self.handle_ready(data)

        elif op == self.HEARTBEAT:
            asyncio.create_task(self.heartbeat(data))

        elif op == self.SESSION_DESCRIPTION:
            await self.handle_description(data)

    async def handle_description(self, data: dict):
        self.secret_key: list[Any] = data.get("secret_key")
        if self.debug:
            log.debug("Finished session description event")
            log.info(f"Gathered Secret Key: {self.secret_key}")

    # def pcm_encode(self, file: str):
    #     out, err = (
    #         ffmpeg
    #         .input(file)
    #         .output('-', format='s16le', acodec='pcm_s16le', ac=2, ar='48k')
    #         .overwrite_output()
    #         .run(capture_stdout=True, capture_stderr=True)
    #     )
    #     return out

    def encode_data(self, data: bytes):
        self.encoder = opuslib.Encoder(
            self.SAMPLING_RATE, self.CHANNELS, opuslib.APPLICATION_AUDIO
        )
        return self.encoder.encode(data, self.SAMPLES_PER_FRAME)

    def get_voice_packet(self, encoded: bytes):
        header = bytearray(12)

        header[0] = 0x80
        header[1] = 0x78
        struct.pack_into(">H", header, 2, self.sequence)
        struct.pack_into(">I", header, 4, self.timestamp)
        struct.pack_into(">I", header, 8, self.SSRC)
        return self.encrypt_xsalsa20_poly1305(header, encoded)

    def encrypt_xsalsa20_poly1305(self, header: bytes, data) -> bytes:
        box = nacl.secret.SecretBox(bytes(self.secret_key))
        nonce = bytearray(24)
        nonce[:12] = header
        return header + box.encrypt(bytes(data), bytes(nonce)).ciphertext

    def checked_add(self, attr: str, value: int, limit: int):
        val = getattr(self, attr)
        if val + value > limit:
            setattr(self, attr, 0)
        else:
            setattr(self, attr, val + value)

    async def send_encode_audio_data(self, data: bytes):
        self.checked_add("sequence", 1, 65535)
        buffer = io.BytesIO(data)
        while True:
            data = buffer.read(self.FRAME_SIZE)
            if len(data) == 0:
                break
            encoded = self.encode_data(data)
            packet = self.get_voice_packet(encoded)
            await self.speak(True)
            self.socket.sendto(packet, (self.endpoint_IP, self.voice_port))

            await asyncio.sleep(self.FRAME_LENGTH / 1000)
        if self.debug:
            log.debug("Send Audio Packet to the voice port")
            log.info(f"Length of data {len(data)}")
        await self.speak(False)
        self.checked_add("timestamp", self.SAMPLES_PER_FRAME, 4294967295)

    async def send_audio_data(self, data: bytes):
        self.checked_add("sequence", 1, 65535)
        buffer = io.BytesIO(data)
        while True:
            data = buffer.read(self.FRAME_SIZE)
            if len(data) == 0:
                break
            packet = self.get_voice_packet(data)
            await self.speak(True)
            self.socket.sendto(packet, (self.endpoint_IP, self.voice_port))
            if self.debug:
                log.debug("Send Audio Packet to the voice port")
                log.info(f"Length of data {len(packet)}")
            await asyncio.sleep(self.FRAME_LENGTH / 1000)
        await self.speak(False)
        self.checked_add("timestamp", self.SAMPLES_PER_FRAME, 4294967295)

    async def play(self, path):
        await self.speak(False)
        if self.debug:
            log.info(f"File Path: {path}")
        if os.path.exists(path):
            if os.path.isfile(path):
                async with aiofiles.open(path, mode="rb") as f:
                    data = await f.read()
                # data = self.pcm_encode(path)
                if path.endswith(".opus"):
                    if self.debug:
                        log.debug("Using non-encoding function for send of voice data")
                    await self.send_audio_data(data)
                else:
                    if self.debug:
                        log.debug("Using encode function for send of voice data")
                    await self.send_encode_audio_data(data)
            else:
                RuntimeError("Path is not a file")
        else:
            RuntimeError("Path does not exist")

    async def speak(self, state):
        payload = {
            "op": 5,
            "d": {
                "speaking": int(state),
                "ssrc": self.SSRC,
                "delay": 0,
            },
        }
        await self.send_json(payload)

    async def send_json(self, payload: dict):
        """Send json to the websocket
        Args:
            payload (dict): Valid payload to send to the gateway
        """
        await self.ws.send(json.dumps(payload))

    async def close(self):
        """Close the connection to websocket"""
        self.alive = False
        await self.ws.close()

    async def identify(self):
        payload = {
            "op": 0,
            "d": {
                "server_id": self.server_id,
                "token": self.token,
                "session_id": self.session_id,
                "user_id": self.bot.user.id,
                "streams": [
                    {"type": "video", "rid": "100", "quality": 100},
                    {"type": "video", "rid": "50", "quality": 50},
                ],
                "video": True,
            },
        }
        await self.send_json(payload)
        if self.debug:
            log.debug("Sent voice identify payload")
            log.info(
                f"Identified on USER: {self.bot.user} SERVER: {self.server_id} TOKEN: {self.token} SESSION_ID: {self.session_id}"
            )

    async def start(self):
        await self.connect()
        await self.identify()
        while self.alive:
            try:
                await self.recv_msg()
            except KeyboardInterrupt:
                log.error("Shutting down")
                await self.close()
            except Exception as e:
                log.error(f"Websocket Unexpectedly closed {e}")
                await self.close()

    async def handle_ready(self, data: dict):
        self.SSRC = data.get("ssrc")
        self.endpoint_IP = data.get("ip")
        self.voice_port = data.get("port")
        await self.ip_discovery()
        if self.debug:
            log.debug("Finished voice ready event")
            log.info(
                f"Gathered SSRC: {self.SSRC} ENDPOINT IP: {self.endpoint_IP} PORT: {self.voice_port}"
            )

    async def heartbeat(self, data: dict):
        interval = data.get("heartbeat_interval") / 1000.0
        while self.alive:
            await asyncio.sleep(interval)
            payload = {"op": 3, "d": time.time()}
            await self.send_json(payload)
            if self.debug:
                log.debug("Sent Heartbeat")

    async def udp_select(self):
        payload = {
            "op": 1,
            "d": {
                "protocol": "udp",
                "codecs": [
                    {
                        "name": "opus",
                        "type": "audio",
                        "priority": 1000,
                        "payload_type": 120,
                    }
                ],
                "data": {"address": self.IP, "port": self.port, "mode": self.mode},
            },
        }
        await self.send_json(payload)
        if self.debug:
            log.debug("Sent UDP select payload")
            log.info(f"Sent on Address {self.IP} and Port {self.port}")

    async def ip_discovery(self):
        self.socket: socket.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setblocking(False)
        packet = bytearray(74)
        struct.pack_into(">H", packet, 0, 1)
        struct.pack_into(">H", packet, 2, 70)
        struct.pack_into(">I", packet, 4, self.SSRC)
        self.socket.sendto(packet, (self.endpoint_IP, self.voice_port))
        if self.debug:
            log.debug("Sent IP discovery packet")
            log.info(
                f"Sent packet to Address: {self.endpoint_IP} and Port: {self.voice_port}"
            )
        while True:
            try:
                data = self.socket.recv(74)
                break
            except:
                continue
        ip_start = 8
        ip_end = data.index(0, ip_start)
        self.IP = data[ip_start:ip_end].decode("ascii")
        self.port = struct.unpack_from(">H", data, len(data) - 2)[0]
        if self.debug:
            log.debug("Received IP discovery response")
            log.info(f"Received Address {self.IP} and Port {self.port}")
        await self.udp_select()



from __future__ import annotations

import asyncio
import time
from time import perf_counter
from typing import TYPE_CHECKING

from aioconsole import aprint

from ..models import (
    Client,
    DMChannel,
    GroupChannel,
    Guild,
    Message,
    TextChannel,
    User,
    VoiceChannel,
)
from ..models.role import Role
from ..utils import logging
from .voice import Voice

if TYPE_CHECKING:
    from ..bot import Bot
    from .http import http

log = logging.getLogger("Event")


class EventHandler:
    """
    Used to handle discord events
    """

    def __init__(self, bot, http, debug=False):
        self._events = {}
        self.http: http = http
        self.bot: Bot = bot
        self.debug: bool = debug

    async def handle_ready(self, data: dict, user: Client, http: http):
        """Handles what happens when the ready event is fired, when the bot first connects

        Args:
            data (dict): JSON data from gateway
            user (Client): The client instance
            http (http): HTTP instance
        """
        self.user = user
        for relationship in data.get("relationships"):
            if relationship.get("type") == 1:
                self.user.friends.append(User(relationship["user"], self.bot, http))

        for channel in data.get("private_channels"):
            if channel.get("type") == 1:
                self.user.private_channels.append(DMChannel(channel, self.bot, http))
            if channel.get("type") == 3:
                self.user.private_channels.append(GroupChannel(channel, self.bot, http))

        for guild in data.get("guilds"):
            await self.handle_guild_create(guild, self.user, http)
        self.bot.session_id = data.get("session_id")
        self.bot.resume_url = data.get("resume_gateway_url")
        # Sends data from ready to the event handler in main.py (if it exists)
        await self.bot.emit("ready", perf_counter() - self.bot.t1)

    async def handle_guild_create(self, data: dict, user: Client, http: http):
        """Handles what happens when a guild is created

        Args:
            data (dict): JSON data from gateway
            user (Client): The client instance
            http (http): HTTP instance
        """
        self.user = user
        guild = Guild(data, self.bot, http)
        self.user.guilds.append(guild)

        # Sends data from ready to the event handler in main.py (if it exists)
        await self.bot.emit("guild_create", guild)

    async def handle_message_create(self, data: dict, user: Client, http: http):
        """Handles what happens when a message is created, or sent
        Args:
            data (dict): JSON data from gateway
            user (Client): The client instance
            http (http): HTTP instance
        """
        self.user = user
        message = Message(data, self.bot, http)
        self.user.messages.append(message)

        # Sends data from ready to the event handler in main.py (if it exists)
        await self.bot.emit("message", message)
        if self.bot.userbot:
            for prefix in self.bot.prefixes:
                # Attempts to invoke the command if has prefix
                if message.content.startswith(prefix):
                    await self.bot.process_commands(message)

        elif message.author.id == self.bot.user.id:
            for prefix in self.bot.prefixes:
                # Attempts to invoke the command if has prefix and from the user
                if message.content.startswith(prefix):
                    await self.bot.process_commands(message)

    async def handle_message_delete(self, data: dict, user: Client, http: http):
        """Handles what happens when a message is deleted. Very little data will be logged if the message is not in the bots cache.

        Args:
            data (dict): JSON data from gateway
            user (Client): The client instance
            http (http): HTTP instance
        """
        self.user = user
        id = data.get("id")
        for message in self.user.messages:
            if message.id == id:
                setattr(message, "deleted_time", time.time())
                await self.bot.emit("message_delete", message)
                self.user.deleted_messages.append(message)
                self.user.messages.remove(message)
        else:  # type: ignore

            class deleted_message:
                def __init__(self, data) -> None:
                    self.tts = data.get("tts")
                    self.references_message = data.get("referenced_message")
                    self.mentions = data.get("mentions")
                    self.author = None
                    self.id = data.get("id")
                    self.flags = data.get("flags")
                    self.embeds = data.get("embeds")
                    self.content = data.get("content")
                    self.components = data.get("components")
                    self.channel_id = data.get("channel_id")
                    self.attachments = data.get("attachments")
                    self.guild_id = data.get("guild_id")
                    self.channel = None
                    self.guild = None
                    self.deleted_time = time.time()

            message = deleted_message(data)
            await self.bot.emit("message_delete", message)

    async def handle_channel_create(self, channel: dict, user: Client, http: http):
        """Handles what happens when a channel is created

        Args:
            channel (dict): JSON data from gateway
            user (Client): The client instance
            http (http): HTTP instance
        """
        self.user = user
        if channel.get("type") == 0 or channel.get("type") not in [1, 2, 3]:
            id = channel.get("guild_id")
            for guild in self.user.guilds:
                if guild.id == id:
                    channel = TextChannel(channel, self.bot, self.http)
                    guild.channels.append(channel)

        elif channel.get("type") != 0 and channel.get("type") == 1:
            self.user.private_channels.append(DMChannel(channel, self.bot, http))

        elif (
            channel.get("type") != 0
            and channel.get("type") != 1
            and channel.get("type") == 2
        ):
            id = channel.get("guild_id")
            for guild in self.user.guilds:
                if guild.id == id:
                    channel = VoiceChannel(channel, self.bot, self.http)
                    guild.channels.append(channel)

        else:
            self.user.private_channels.append(GroupChannel(channel, self.bot, http))

        # Sends data from ready to the event handler in main.py (if it exists)
        await self.bot.emit("channel_create", channel)

    async def handle_guild_member_list_update(
        self, data: dict, user: Client, http: http
    ):
        """Handles what happens when a member chunk payload is received

        Args:
            data (dict): JSON data from gateway
            user (Client): The client instance
            http (http): HTTP instance
        """
        self.user = user
        ops = []
        data = data["ops"]
        for main in data:
            op = main["op"]
            ops.append(op)

        for main in data:
            if ("INSERT" and "DELETE") in ops:
                try:
                    person = User(main["item"]["member"]["user"], self.bot, self.http)
                    await self.bot.emit(
                        "member_chunk",
                        members=None,
                        other=None,
                        join=None,
                        leave=person,
                    )
                except:
                    continue
            elif "INSERT" in ops:
                try:
                    person = User(main["item"]["member"]["user"], self.bot, self.http)
                    await self.bot.emit(
                        "member_chunk",
                        members=None,
                        other=None,
                        join=person,
                        leave=None,
                    )
                except:
                    continue
            elif "UPDATE" in ops:
                try:
                    person = User(main["item"]["member"]["user"], self.bot, self.http)
                    await self.bot.emit(
                        "member_chunk",
                        members=None,
                        other=person,
                        join=None,
                        leave=None,
                    )
                except:
                    continue
            elif "SYNC" in ops:
                users = []
                for item in main["items"]:
                    try:
                        users.append(User(item["member"]["user"], self.bot, self.http))
                    except:
                        continue
                await self.bot.emit(
                    "member_chunk", members=users, other=None, join=None, leave=None
                )

    async def handle_channel_delete(self, data, user: Client, http: http):
        """Handles what happens when a channel is deleted

        Args:
            data (dict): JSON data from gateway
            user (Client): The client instance
            http (http): HTTP instance
        """
        self.user = user
        id = data.get("id")
        for channel in self.user.private_channels:
            if channel.id == id:
                await self.bot.emit("channel_delete", channel)
                self.user.private_channels.remove(channel)
                return

        guild_id = data.get("guild_id")
        if guild_id != None:
            for guild in self.user.guilds:
                if guild_id == guild.id:
                    for channel in guild.channels:
                        if channel.id == id:
                            await self.bot.emit("channel_delete", channel)
                            guild.channels.remove(channel)
                            return
        else:
            for guild in self.user.guilds:
                for channel in guild.channels:
                    if channel.id == id:
                        await self.bot.emit("channel_delete", channel)
                        guild.channels.remove(channel)
                        return

    async def handle_guild_role_create(self, role, user: Client, http: http):
        """Handles what happens when a role is created

        Args:
            data (dict): JSON data from gateway
            user (Client): The client instance
            http (http): HTTP instance
        """
        self.user = user

        for guild in self.user.guilds:
            if role.get("guild_id") == guild.id:
                role = Role(role, self.bot, self.http, guild_id=guild.id)
                guild.roles.append(role)

        await self.bot.emit("role_create", role)

    async def handle_guild_role_delete(self, role: dict, user: Client, http: http):
        """Handles what happens when a role is deleted

        Args:
            data (dict): JSON data from gateway
            user (Client): The client instance
            http (http): HTTP instance
        """
        self.user = user

        for guild in self.user.guilds:
            if role.get("guild_id") == guild.id:
                for role in guild.roles:
                    if role.get("id") == role.id:
                        await self.bot.emit("role_delete", role)
                        guild.roles.remove(role)
                        return

    async def handle_call_update(self, data: dict, user: Client, http: http):
        channel = self.bot.get_channel(data["channel_id"])
        region = data.get("region")
        if isinstance(channel, DMChannel):
            users = channel.recipient
        elif isinstance(channel, GroupChannel):
            users = channel.recipients

        await self.bot.emit("call_update", channel, users, region)

    async def handle_call_create(self, data: dict, user: Client, http: http):
        channel = self.bot.get_channel(data["channel_id"])
        region = data.get("region")
        if isinstance(channel, DMChannel):
            users = channel.recipient
        elif isinstance(channel, GroupChannel):
            users = channel.recipients

        await self.bot.emit("call_create", channel, users, region)

    async def handle_call_delete(self, data: dict, user: Client, http: http):
        channel = self.bot.get_channel(data["channel_id"])
        await self.bot.emit("call_delete", channel)

    async def handle_voice_state_update(self, data: dict, user: Client, http: http):
        """Handles the voice state updating

        Args:
            data (dict): JSON data from gateway
            user (Client): The client instance
            http (http): HTTP instance
        """
        if data["channel_id"] != None:
            self.session_id = data["session_id"]

    async def handle_presence_update(self, data: dict, user: Client, http: http):
        """Handles the presence updating

        Args:
            data (dict): JSON data from gateway
            user (Client): The client instance
            http (http): HTTP instance
        """
        LISTENING = 2
        CUSTOM = 4

        last_modified = data.get("last_modified")
        status = data.get("status")
        check = data.get("user").get("username")
        if check is None:
            user = data.get("user").get("id")

        else:
            user = User(data.get("user"), self.bot, self.http)
        client_status = data.get("client_status")
        activity = data.get("activities")
        activities = []
        if activity != None:
            PLAYING = 0
            STREAMING = 1
            WATCHING = 3
            for activity in activity:
                type = activity.get("type")
                if type == PLAYING:
                    type = "PLAYING"
                elif type == STREAMING:
                    type = "STREAMING"
                elif type == WATCHING:
                    type == "WATCHING"
                elif type == 4:
                    type = "CUSTOM"
                activities.append({"Type": type, "Name": activity.get("name")})

        await self.bot.emit(
            "presence_update", user, status, last_modified, client_status, activities
        )

    async def handle_voice_server_update(self, data: dict, user: Client, http: http):
        """Handles the voice server updating

        Args:
            data (dict): JSON data from gateway
            user (Client): The client instance
            http (http): HTTP instance
        """
        self.token = data["token"]
        self.endpoint = data["endpoint"]
        if data["guild_id"] is None:
            self.server_id = data["channel_id"]
        else:
            self.server_id = data["guild_id"]
        await asyncio.sleep(1)
        self.voice = Voice(
            self.session_id,
            self.token,
            self.endpoint,
            self.server_id,
            self.bot,
            debug=self.debug,
        )
        await self.voice_start(self.voice)

    async def voice_start(self, voice: Voice):
        asyncio.create_task(voice.start())
        setattr(self.bot, "voice", self.voice)
        if self.debug:
            log.debug("Voice attribute created")
            log.info(
                f"Created voice attribute with Session ID: {self.session_id} Endpoint: {self.endpoint}"
            )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .channel import Voiceable
    from .guild import Guild
    from .message import Message
    from .user import User


class Client:
    """Client Object"""

    def __init__(self, UserPayload: dict) -> None:
        self.guilds: list[Guild] = []
        self.private_channels: list[Voiceable] = []
        self.friends: list[User] = []
        self.messages: list[Message] = []
        self.deleted_messages: list[Message] = []
        self._update(UserPayload)

    def __str__(self) -> str:
        return f"""{self.name}#{self.discriminator}"""

    def __eq__(self, other):
        return self.id == other.id

    def _update(self, data):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """
        self.name = data.get("username")
        self.id = data.get("id")
        self.discriminator = data.get("discriminator")
        self.avatar = data.get("avatar")
        self.banner = data.get("banner")
        self.bio = data.get("bio")
        self.email = data.get("email")
        self.phone = data.get("phone")
        self.accent_colour = data.get("accent_color")
        self.public_flags = data.get("public_flags")
        self.bot_acc = data.get("bot")
        self.system = data.get("system")



from __future__ import annotations

import asyncio
import json
import os
import time
import zlib
from traceback import format_exception

import requests
import websockets

from selfcord.models.client import Client

from ..utils import logging
from .errors import ReconnectWebsocket
from .events import EventHandler

log = logging.getLogger("Gateway")


class Activity:
    @staticmethod
    def Game(
        name: str,
        details: str,
        state: str,
        buttons: dict,
        application_id: str,
        key: str,
    ) -> dict[str, int]:
        """Method to generate activity dict for the "Playing ..." payload

        Args:
            name (str): Name of the activity
            details (str): Details of the activity
            state (str): State of the activity
            buttons (dict): Buttons for the activity.
            Example:
                { "My Website": "https://google.com"}
            application_id (str): Application ID
            key (str): Key for the large image

        Returns:
            dict[str, int]: Dict for the activity object for payload
        """
        type = 0
        button_urls = list(buttons.values())
        buttons: list = list(buttons.keys())
        req = requests.get(
            f"https://discordapp.com/api/oauth2/applications/{application_id}/assets"
        )
        for item in req.json():
            if item["name"] == key:
                key = item["id"]

        return (
            {
                "name": name,
                "details": details,
                "state": state,
                "application_id": application_id,
                "assets": {
                    "large_image": key,
                },
                "type": type,
                "created_at": int(time.time()),
            }
            if not button_urls or not buttons
            else {
                "name": name,
                "details": details,
                "type": type,
                "buttons": buttons,
                "state": state,
                "application_id": application_id,
                "assets": {
                    "large_image": key,
                },
                "metadata": {
                    "button_urls": button_urls,
                },
                "created_at": int(time.time()),
            }
        )

    @staticmethod
    def Stream(
        name: str,
        details: str,
        state: str,
        url: str,
        buttons: dict,
        application_id: str,
        key: str,
    ) -> dict[str, int]:
        """Method to generate activity dict for the "Streaming ..." payload

        Args:
            name (str): Name of the activity
            details (str): Details of the activity
            state (str): State of the activitiy
            url (str): URL for streaming
            buttons (dict): Buttons for the activity.
            Example:
                { "My Website": "https://google.com"}
            application_id (str): Application ID
            key (str): Key for the large image

        Returns:
            dict[str, int]: Dict for the activity object for payload
        """
        type = 1
        button_urls = list(buttons.values())
        buttons: list = list(buttons.keys())
        req = requests.get(
            f"https://discordapp.com/api/oauth2/applications/{application_id}/assets"
        )
        for item in req.json():
            if item["name"] == key:
                key = item["id"]

        return (
            {
                "name": name,
                "details": details,
                "type": type,
                "state": state,
                "application_id": application_id,
                "assets": {
                    "large_image": key,
                },
                "created_at": int(time.time()),
                "url": url,
            }
            if not button_urls or not buttons
            else {
                "name": name,
                "details": details,
                "type": type,
                "buttons": buttons,
                "url": url,
                "state": state,
                "application_id": application_id,
                "assets": {
                    "large_image": key,
                },
                "metadata": {
                    "button_urls": button_urls,
                },
                "created_at": int(time.time()),
            }
        )

    @staticmethod
    def Listen(
        name: str,
        details: str,
        state: str,
        buttons: dict,
        application_id: str,
        key: str,
    ) -> dict[str, int]:
        """Method to generate activity dict for the "Listening ..." payload

        Args:
            name (str): Name of the activity
            details (str): Details of the activity
            state (str): State of the activity
            buttons (dict): Buttons for the activity.
            Example:
                { "My Website": "https://google.com"}
            application_id (str): Application ID
            key (str): Key for the large image

        Returns:
            dict[str, int]: Dict for the activity object for payload
        """
        type = 2
        button_urls = list(buttons.values())
        buttons: list = list(buttons.keys())
        req = requests.get(
            f"https://discordapp.com/api/oauth2/applications/{application_id}/assets"
        )
        for item in req.json():
            if item["name"] == key:
                key = item["id"]

        return (
            {
                "name": name,
                "details": details,
                "state": state,
                "application_id": application_id,
                "assets": {
                    "large_image": key,
                },
                "type": type,
                "created_at": int(time.time()),
            }
            if not button_urls or not buttons
            else {
                "name": name,
                "details": details,
                "type": type,
                "buttons": buttons,
                "state": state,
                "application_id": application_id,
                "assets": {
                    "large_image": key,
                },
                "metadata": {
                    "button_urls": button_urls,
                },
                "created_at": int(time.time()),
            }
        )

    @staticmethod
    def Watch(
        name: str,
        details: str,
        state: str,
        buttons: dict,
        application_id: str,
        key: str,
    ) -> dict[str, int]:
        """Method to generate activity dict for the "Watching ..." payload

        Args:
            name (str): Name of the activity
            details (str): Details of the activity
            state (str): State of the activity
            buttons (dict): Buttons for the activity.
            Example:
                { "My Website": "https://google.com"}
            application_id (str): Application ID
            key (str): Key for the large image

        Returns:
            dict[str, int]: Dict for the activity object for payload
        """
        type = 3

        button_urls = list(buttons.values())
        buttons: list = list(buttons.keys())

        req = requests.get(
            f"https://discordapp.com/api/oauth2/applications/{application_id}/assets"
        )
        for item in req.json():
            if item["name"] == key:
                key = item["id"]

        return (
            {
                "name": name,
                "details": details,
                "state": state,
                "application_id": application_id,
                "assets": {
                    "large_image": key,
                },
                "type": type,
                "created_at": int(time.time()),
            }
            if not button_urls or not buttons
            else {
                "name": name,
                "details": details,
                "type": type,
                "buttons": buttons,
                "state": state,
                "application_id": application_id,
                "assets": {
                    "large_image": key,
                },
                "metadata": {
                    "button_urls": button_urls,
                },
                "created_at": int(time.time()),
            }
        )


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from zlib import _Decompress

    from .http import http


class gateway:
    """Discord Gateway instance, used to initialise gateway connections"""

    DISPATCH = 0
    HEARTBEAT = 1
    IDENTIFY = 2
    PRESENCE = 3
    VOICE_STATE = 4
    VOICE_PING = 5
    RESUME = 6
    RECONNECT = 7
    REQUEST_MEMBERS = 8
    INVALIDATE_SESSION = 9
    HELLO = 10
    HEARTBEAT_ACK = 11
    GUILD_SYNC = 12

    def __init__(self, http: http, debug: bool = False):
        self.debug: bool = debug
        self.http: http = http
        self.zlib: _Decompress = zlib.decompressobj()
        self.zlib_suffix: bytes = b"\x00\x00\xff\xff"
        self.last_ack: float = time.perf_counter()
        self.last_send: float = time.perf_counter()
        self.latency = float("inf")
        self.resume_url = None
        self.alive = False
        self.fired = False

    async def recv_msg(self):
        """
        Receives Message from gateway, encodes as json and does things depending on op code
        """
        item = await self.ws.recv()
        buffer = bytearray()
        buffer.extend(item)
        if len(item) < 4 or item[-4:] != self.zlib_suffix:
            return

        if item:
            if self.fired:
                log.info(f"{item}")
            try:
                item = self.zlib.decompress(item)
            except Exception as e:
                log.error(f"Could not decompress\n{e}")
            item = json.loads(item)  # Get json message from gateway

            op = item.get("op")  # Op code
            data = item.get("d")  # Data
            event = item.get("t")  # The event
            s = item.get("s")
            if op == self.RECONNECT:
                await self.close()
                await self.reconnect(s)
                self.fired = True
                log.error("Reconnect websocket")

            elif op == self.INVALIDATE_SESSION:
                if data:
                    await self.close()
                    log.error("Session Invalidated")

            elif op == self.HELLO:
                # Begins heartbeat and sends identify if this op is received
                interval = data["heartbeat_interval"] / 1000.0
                await self.identify()
                asyncio.create_task(self.heartbeat(interval))

            elif op == self.HEARTBEAT_ACK:
                await self.heartbeat_ack()

            elif op == self.DISPATCH:
                # If op is 0 it signifies a regular gateway event
                # These events are discord events like message_create, role_create whatever.

                handle = f"handle_{event.lower()}"

                if hasattr(
                    self.handler, handle
                ):  # If the event handler exists, so e.g handle_ready
                    method = getattr(self.handler, handle)

                    val = await asyncio.gather(
                        asyncio.create_task(method(data, self.user, self.http)),
                        return_exceptions=True,
                    )  # A background task is created to run the handler
                    for item in val:
                        if item is None:
                            break
                        else:
                            await self.bot.emit("error", item)

                    # asyncio.create_task(method(data, self.user, self.http))
                # Handlers are all situated in events.py

    def roundup(self, n):
        import math

        return int(math.ceil(n / 100.0)) * 100

    def chunks(self, lst, n):
        for i in range(0, len(lst), 1):
            if len(lst[: i + 1]) > 3:
                for i in range(i, len(lst), n):
                    yield lst[i : i + n]
                break

            yield lst[: i + 1]

    async def change_presence(self, status: str, afk: bool, activity: dict):
        """Change the clients current presence

        Args:
            status (str): online, offline or dnd
            afk (bool): Whether client is set as AFK
            activity (Activity): Activity object
        """
        payload = {
            "op": 3,
            "d": {
                "since": time.time(),
                "activities": [activity],
                "status": status.lower(),
                "afk": afk,
            },
        }

        await self.send_json(payload)
        if self.debug:
            log.debug("Began rich presence")
            log.info(f"{activity}")

    async def lazy_chunk(self, guild_id: str, channel_id: str, amount: int):
        """Sends lazy guild request to gather current online members

        Args:
            guild_id (str): The guild id specified
            channel_id (str): The channel id specified
        """

        ranges = []

        for i in range(0, amount, 100):
            ranges.append(
                [i, self.roundup(i + (amount - i)) - 1]
            ) if i + 99 > amount else ranges.append([i, i + 99])

        for item in self.chunks(ranges, 3):
            payload = {
                "op": 14,
                "d": {
                    "guild_id": guild_id,
                    "typing": True,
                    "channels": {channel_id: item},
                },
            }

            await self.send_json(payload)
            await asyncio.sleep(2.0)

        if self.debug:
            log.debug("Finished guild lazy chunking")
            log.info(f"Subscription to GUILD: {guild_id} with CHANNEL: {channel_id}")

    async def send_json(self, payload: dict):
        """Send json to the gateway

        Args:
            payload (dict): Valid payload to send to the gateway
        """
        await self.ws.send(json.dumps(payload))

    async def reconnect(self, seq: int):
        """Reconnect to discord gateway"""
        self.ws = await websockets.connect(
            f"{self.bot.resume_url}?encoding=json&v=9&compress=zlib-stream"
        )
        self.alive = True
        payload = {
            "op": 6,
            "d": {"token": self.token, "session_id": self.bot.session_id, "seq": seq},
        }
        await self.send_json(payload)

    async def connect(self):
        """Connect to discord gateway"""
        self.ws = await websockets.connect(
            "wss://gateway.discord.gg/?encoding=json&v=9&compress=zlib-stream",
            origin="https://discord.com",
            max_size=None,
        )
        self.alive = True
        if self.debug:
            log.debug("Established connection to Discord Gateway")

    async def close(self):
        """Close the connection to discord gateway"""
        self.alive = False
        await self.ws.close()
        if self.debug:
            log.debug("Closed connection to discord gateway")

    async def identify(self):
        """Identify to gateway, uses amazing mobile client spoof"""
        payload = {
            "op": 2,
            "d": {
                "token": self.token,
                "properties": {
                    "$os": "android",
                    "$browser": "Discord Android",
                    "$device": "Discord Android",
                    "$referrer": "",
                    "$referring_domain": "",
                },
            },
        }
        await self.send_json(payload)
        if self.debug:
            log.debug("Sent identify payload")
            log.info(f"Idenitified with {self.token} under Discord Android")

    async def heartbeat(self, interval):
        """Heartbeat for gateway to maintain connection

        Args:
            interval (int): Interval between sends
        """
        log.info(f"Hearbeat loop has began with the interval of {interval} seconds!")
        heartbeatJSON = {"op": 1, "d": time.time()}
        while True:
            await asyncio.sleep(interval)
            await self.send_json(heartbeatJSON)
            self.last_send = time.perf_counter()
            if self.debug:
                log.debug("Sent heartbeat")
                log.info(f"Delay since last heartbeat {self.latency:0.2f}ms")

    async def heartbeat_ack(self):
        """Whenever heartbeat ack is sent, logs the time between last send of heartbeat json and receive of the ack"""
        self.last_ack = time.perf_counter()
        self.latency = self.last_ack - self.last_send

    async def start(self, token: str, user: Client, bot):
        """Start discord gateway connection

        Args:
            token (str): User token
            user (Client): User client
            bot (_type_): Bot class
        """
        self.handler = EventHandler(bot, self.http, self.debug)
        self.bot = bot

        await self.bot.inbuilt_commands()  # In built commands very cool

        self.user = user
        self.token = token

        await self.connect()
        while self.alive:
            try:
                await self.recv_msg()
            except KeyboardInterrupt:
                log.critical("Shutting down...")
                await self.close()
            except Exception as e:
                error = "".join(format_exception(e, e, e.__traceback__))
                await self.bot.emit("error", error)
                if self.debug:
                    log.error(f"Websocket Unexpectedly closed\n{error}")
                await self.close()

    async def video_call(self, channel: str, guild=None):
        """Initiates a discord video call

        Args:
            channel (str): Channel ID
            guild (str, optional): Guild ID. Defaults to None.
        """
        payload = {
            "op": 4,
            "d": {
                "guild_id": guild,
                "channel_id": channel,
                "preferred_region": "rotterdam",
                "self_mute": False,
                "self_deaf": False,
                "self_video": True,
            },
        }
        await self.send_json(payload)
        if self.debug:
            log.debug("Initiated video call")
            log.info(f"Began video call to GUILD: {guild} and CHANNEL: {channel}")
        if guild is None:
            await self.http.request(
                method="post",
                endpoint=f"/channels/{channel}/call/ring",
                json={"recipients": None},
            )
            if self.debug:
                log.debug("Sent Ring")

    async def call(self, channel: str, guild=None):
        """Initiates a discord call

        Args:
            channel (str): Channel ID
            guild (str, optional): Guild ID. Defaults to None.
        """
        payload = {
            "op": 4,
            "d": {
                "guild_id": guild,
                "channel_id": channel,
                "preferred_region": "rotterdam",
                "self_mute": False,
                "self_deaf": False,
                "self_video": False,
            },
        }
        await self.send_json(payload)
        if self.debug:
            log.debug("Initiated call")
            log.info(f"Began call to GUILD: {guild} and CHANNEL: {channel}")
        if guild is None:
            await self.http.request(
                method="post",
                endpoint=f"/channels/{channel}/call/ring",
                json={"recipients": None},
            )
            if self.debug:
                log.debug("Sent Ring")

    async def stream_call(self, channel: str, guild=None):
        """Initiates a discord stream call

        Args:
            channel (str): Channel ID
            guild (str, optional): Guild ID. Defaults to None.
        """
        await self.call(channel, guild)
        type = "guild" if guild != None else "call"
        payload = {
            "op": 18,
            "d": {
                "guild_id": guild,
                "channel_id": channel,
                "preferred_region": "rotterdam",
                "type": type,
            },
        }
        await self.send_json(payload)
        if self.debug:
            log.debug("Initiated call")
            log.info(f"Began call to GUILD: {guild} and CHANNEL: {channel}")
        if guild is None:
            await self.http.request(
                method="post",
                endpoint=f"/channels/{channel}/call/ring",
                json={"recipients": None},
            )

            payload = {
                "op": 22,
                "d": {"stream_key": f"call:{channel}:{self.user.id}", "paused": False},
            }
            if self.debug:
                log.debug("Initiated streaming")
                log.info(f"Began stream to CHANNEL: {channel}")
        else:
            payload = {
                "op": 22,
                "d": {
                    "stream_key": f"guild:{guild}:{channel}:{self.user.id}",
                    "paused": False,
                },
            }
            if self.debug:
                log.debug("Initiated streaming")
                log.info(f"Began stream to GUILD: {guild} and CHANNEL: {channel}")

        await self.send_json(payload)

    async def leave_call(self):
        """Leaves a discord call"""
        payload = {
            "op": 4,
            "d": {
                "guild_id": None,
                "channel_id": None,
                "self_mute": False,
                "self_deaf": False,
                "self_video": False,
            },
        }
        await self.send_json(payload)
        if self.debug:
            log.debug("Left Voice call")
        if hasattr(self.bot, "voice"):
            await self.bot.voice.close()
            delattr(self.bot, "voice")
        if self.debug:
            log.info("Voice attribute for bot has been deleted")



from __future__ import annotations

import random
from datetime import datetime, timedelta, timezone
from itertools import zip_longest
from typing import TYPE_CHECKING

from .channel import Category, Messageable, TextChannel, VoiceChannel
from .emoji import Emoji
from .role import Role
from .user import User

if TYPE_CHECKING:
    from ..api import http
    from ..bot import Bot


class Guild:
    """Guild Object"""

    TEXTCHANNEL = 0
    VOICECHANNEL = 2
    CATEGORY = 4
    GUILD_ANNOUNCEMENT = 5
    ANNOUNCEMENT_THREAD = 10
    PUBLIC_THREAD = 11
    PRIVATE_THREAD = 12
    GUILD_STAGE_VOICE = 13
    GUILD_DIRECTORY = 14
    GUILD_FORUM = 15

    def __init__(self, data: dict, bot: Bot, http: http) -> None:
        self.roles: list[Role] = []
        self.emojis: list[Emoji] = []
        self.members: list[User] = []
        self.channels: list[Messageable] = []
        self.http: http = http
        self.bot: Bot = bot
        self._update(data)

    def __str__(self) -> str:
        return f"{self.name}"

    def __eq__(self, other):
        return self.id == other.id

    def _update(self, data):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """
        self.id = data.get("id")
        self.name = data.get("name")
        self.icon = data.get("icon")
        self.icon_url = (
            f"https://cdn.discordapp.com/icons/{self.id}/{self.icon}.webp?size=96"
        )
        self.region = data.get("region")
        self.splash = data.get("splash")
        self.mfa_level = data.get("mfa_level")
        self.features = data.get("features")
        self.member_count = data.get("member_count")
        self.unavailable = data.get("unavailable")
        self.verification_level = data.get("verification_level")
        self.explicit_content_filter = data.get("explicit_content_filter")
        self.owner_id = data.get("owner_id")
        members, channels, roles, emojis = (data.get("members") if data.get("members") is not None else [], data.get("channels") if data.get("channels") is not None else [],data.get("roles") if data.get("roles") is not None else [], data.get("emojis") if data.get("emojis") is not None else [])
        
        for member, channel, role, emoji in zip_longest(
            members,
            channels,
            roles,
            emojis
        ):
            if member != None:
                user = User(member, self.bot, self.http)

                self.members.append(user)

            if channel != None:
                type = channel.get("type")
                if type == self.TEXTCHANNEL or type not in [
                    self.VOICECHANNEL,
                    self.CATEGORY,
                ]:
                    channel = TextChannel(channel, self.bot, self.http)
                elif type == self.VOICECHANNEL:
                    channel = VoiceChannel(channel, self.bot, self.http)
                else:
                    channel = Category(channel, self.bot, self.http)
                channel.guild_id = self.id
                self.channels.append(channel)
            if role != None:
                role = Role(role, self.bot, self.http, guild_id=self.id)
                self.roles.append(role)

            if emoji != None:
                emoji = Emoji(emoji, self.bot, self.http)
                emoji.guild_id = self.id
                self.emojis.append(emoji)

    async def ban(self, user_id: str):
        """Bans a user from the guild

        Args:
            user_id (str): User ID specified to ban
        """
        await self.http.request(
            method="put",
            endpoint=f"/guilds/{self.id}/bans/{user_id}",
            json={"delete_message_days": "7"},
        )

    async def kick(self, user_id: str):
        """Kicks a user from the guild

        Args:
            user_id (str): User ID specified to kick
        """
        await self.http.request(
            method="delete", endpoint=f"/guilds/{self.id}/members/{user_id}"
        )

    def utc_now(self):
        return datetime.now(timezone.utc)

    async def timeout(
        self, user_id: str, hours: int = 0, mins: int = 0, seconds: int = 0
    ):
        """Timeouts a user in the guild

        Args:
            user_id (str): User ID specified to timeout
        """
        duration = self.utc_now() + timedelta(
            hours=hours, minutes=mins, seconds=seconds
        )
        await self.http.request(
            method="patch",
            endpoint=f"/guilds/{self.id}/members/{user_id}",
            json={"communication_disabled_until": str(duration)},
        )

    async def txt_channel_create(self, name: str, parent_id: str = None):
        """Creates a Text Channel in the guild

        Args:
            name (str): Name of the channel
            parent_id (str, optional): ID of the category. Defaults to None.
        """
        payload = {"name": name, "permission_overwrites": [], "type": 0}
        if parent_id != None:
            payload["parent_id"] = parent_id

        channel = await self.http.request(
            method="post", endpoint=f"/guilds/{self.id}/channels", json=payload
        )
        return TextChannel(channel, self.bot, self.http)

    async def vc_channel_create(self, name: str):
        """Creates a voice channel in the guild

        Args:
            name (str): Name of the channel
        """
        channel = await self.http.request(
            method="post",
            endpoint=f"/guilds/{self.id}/channels",
            json={"name": f"{name}", "permission_overwrites": [], "type": 2},
        )
        return VoiceChannel(channel, self.bot, self.http)

    async def role_create(self, name: str):
        """Creates a role in the guild

        Args:
            name (str): Name of the role
        """
        role = await self.http.request(
            method="post", endpoint=f"/guilds/{self.id}/roles", json={"name": f"{name}"}
        )
        return Role(role, self.bot, self.http, guild_id=self.id)

    async def category_channel_create(self, name: str):
        """Creates a category in the guild

        Args:
            name (str): Name of the category
        """
        channel = await self.http.request(
            method="post",
            endpoint=f"/guilds/{self.id}/channels",
            json={"name": f"{name}", "permission_overwrites": [], "type": 4},
        )
        return Category(channel, self.bot, self.http)

    async def emoji_create(self, name: str, image_url: str):
        """Creates an emoji in the guild

        Args:
            name (str): Name of the emoji
            image_url (str): URL for an image
        """
        image = await self.http.encode_image(image_url)
        emoji = await self.http.request(
            method="post",
            endpoint=f"/guilds/{self.id}/emojis",
            json={"name": f"{name}", "image": image},
        )
        return Emoji(emoji, self.bot, self.http)

    async def get_members(self, channel_id: str):
        """Get guild members for a guild via chunking

        Args:
            channel_id (str): Channel ID to chunk from
        """
        await self.bot.gateway.lazy_chunk(self.id, channel_id, self.member_count)

    async def edit(
        self,
        name: str = None,
        icon_url: str = None,
        banner_url: str = None,
        description: str = None,
    ):
        """Edits attributes for a guild

        Args:
            name (str, optional): Name of the guild. Defaults to None.
            icon_url (str, optional): Image URL for Icon. Defaults to None.
            banner_url (str, optional): Image URL for Banner. Defaults to None.
            description (str, optional): Description of the guild. Defaults to None.
        """
        fields = {}
        if name != None:
            fields["name"] = name

        if description != None:
            fields["description"] = description

        if icon_url != None:
            data = await self.http.encode_image(icon_url)
            fields["icon"] = data

        if banner_url != None:
            data = await self.http.encode_image(banner_url)
            fields["banner"] = data

        await self.http.request(
            method="patch",
            endpoint=f"/guilds/{self.id}",
            headers={
                "origin": "https://discord.com",
                "referer": f"https://discord.com/channels/{self.id}/{random.choice(self.channels)}",
            },
            json=fields,
        )

    async def delete(self):
        """Deletes the Guild Object"""
        await self.http.request(method="delete", endpoint=f"/guilds/{self.id}")



from __future__ import annotations

import asyncio
import time
import urllib
from itertools import zip_longest

from .user import User


class Attachment:
    def __init__(self, data) -> None:
        self.proxy_url = data.get("proxy_url")
        self.url = data.get("url")
        self.name = data.get("filename")
        self.size = data.get("size")
        self.id = data.get("id")


class Message:
    """Message Object"""

    def __init__(self, data, bot, http) -> None:
        self.bot = bot
        self.channel = None
        self.guild = None
        self.http = http
        self._update(data)

    def __str__(self) -> str:
        return f"{self.content}"

    def __eq__(self, other):
        return self.id == other.id

    def _update(self, data):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """
        self.tts = data.get("tts")
        self.referenced_message = data.get("referenced_message")
        self.author = User(data.get("author"), self.bot, self.http)
        self.id = data.get("id")
        self.flags = data.get("flags")
        self.embeds = data.get("embeds")
        self.content = data.get("content")
        self.components = data.get("components")
        self.timestamp = time.time()
        self.channel_id = data.get("channel_id")

        attachments = data.get("attachments")
        self.attachments = []
        self.mentions = []
        for mention, atch in zip_longest(data.get("mentions"), attachments):
            if atch is not None:
                self.attachments.append(Attachment(atch))
            if mention is not None:
                self.mentions.append(User(mention, self.bot, self.http))
        self.guild_id = data.get("guild_id")

        self.channel = self.bot.get_channel(self.channel_id)

        self.guild = self.bot.get_guild(self.guild_id)

    async def delete(self):
        """Delete the Message Object"""
        await self.http.request(
            method="delete", endpoint=f"/channels/{self.channel_id}/messages/{self.id}"
        )

    async def edit(self, content: str, file_paths: list[str] = []) -> Message:
        """Edits the specified message

        Args:
            content (str): Content to edit message to.
        """
        json = { "content":  content }
        if file_paths != []:
            vals = await self.channel.upload_image(file_paths)
            json |= {"attachments" : vals}

        if self.guild_id is None:
            resp = await self.http.request(
                method="patch",
                endpoint=f"/channels/{self.channel_id}/messages/{self.id}",
                headers={
                    "origin": "https://discord.com",
                    "referer": f"https://discord.com/channels/{self.channel_id}",
                },
                json=json,
            )
        else:
            resp = await self.http.request(
                method="patch",
                endpoint=f"/channels/{self.channel_id}/messages/{self.id}",
                headers={
                    "origin": "https://discord.com",
                    "referer": f"https://discord.com/channels/{self.channel_id}/{self.guild_id}",
                },
                json=json,
            )
            resp.update({"guild_id": self.guild_id})
        return Message(resp, self.bot, self.http)

    async def react(self, emoji: str):
        """React to a message with an emoji

        Args:
            emoji (str): The emoji
        """
        raw_reaction = urllib.parse.urlencode({"emoji": emoji}).split("emoji=")[1]
        await self.http.request(
            method="put",
            endpoint=f"/channels/{self.channel_id}/messages/{self.id}/reactions/{raw_reaction}/%40me?location=Message&burst=false",
            headers={"referer": f"https://discord.com/channels/@me/{self.channel_id}"},
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from .permission import Permission

if TYPE_CHECKING:
    from ..api.http import http
    from ..bot import Bot


class Role:
    """Role Object"""

    def __init__(self, data: dict, bot: Bot, http: http, **kwargs) -> None:
        self.guild_id = kwargs.get("guild_id")
        self.http: http = http
        self.bot: Bot = bot
        self._update(data)

    def __eq__(self, other):
        return self.id == other.id

    def _update(self, data):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """
        role = data.get("role")

        self.id = data.get("id") if role is None else role.get("id")
        self.permissions = (
            Permission(int(data.get("permissions"))).permissions
            if role is None
            else Permission(int(role.get("permissions"))).permissions
        )
        self.name = data.get("name") if role is None else role.get("name")
        self.mentionable = (
            data.get("mentionable") if role is None else role.get("mentionable")
        )
        self.managed = data.get("managed") if role is None else role.get("managed")
        self.icon = data.get("icon") if role is None else role.get("icon")
        self.flags = data.get("flags") if role is None else role.get("flags")

        self.color = data.get("color") if role is None else role.get("color")
        self.hoist = data.get("hoist") if role is None else role.get("hoist")
        if self.guild_id is None:
            self.guild_id = data.get("guild_id")

    def __str__(self) -> str:
        return f"{self.name}"

    async def delete(self):
        """Delete the Role Object"""
        await self.http.request(
            method="delete", endpoint=f"/guilds/{self.guild_id}/roles/{self.id}"
        )
        del self



from __future__ import annotations


class DiscordException(Exception):
    pass


class LoginFailure(DiscordException):
    def __init__(self, message: dict, status: int) -> None:
        self.message = message
        self.status = status


class ReconnectWebsocket(DiscordException):
    def __init__(self, message: str) -> None:
        self.message = message


class RuntimeError(DiscordException):
    def __init__(self, message: str) -> None:
        self.message = message


class Funnu(DiscordException):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



from __future__ import annotations

import json
import random
import string
import time
from itertools import zip_longest
from typing import TYPE_CHECKING

from aioconsole import aprint

if TYPE_CHECKING:
    from selfcord.api.http import http
    from selfcord.bot import Bot


class Search:
    """Search object returned from Application command search"""

    def __init__(
        self, application_commands: dict, prev_cursor: str, next_cursor: str
    ) -> None:
        self.commands: list[SlashCommand] = [
            SlashCommand(command) for command in application_commands
        ]
        self.prev_cursor: str = prev_cursor
        self.next_cursor: str = next_cursor


class Option:
    def __init__(self, data: dict) -> None:
        self.name = data["name"]
        self.type = data["type"]
        self.description = data["description"]
        self.options: list[Option] | None = (
            [Option(option) for option in data.get("options")]
            if data.get("options") is not None
            else None
        )
        self.required = data.get("required")
        self.raw_data = data
        self.value: str = None

    def __str__(self):
        return f"{self.name}"

    def __iter__(self):
        if self.options is not None:
            yield from self.options
        else:
            yield None


class SlashCommand:
    def __init__(self, data: dict) -> None:
        self.id: str = data["id"]
        self.name: str = data["name"]
        self.type: int = data["type"]
        self.version: str = data["version"]
        self.raw_data: str = data
        self.options: list[Option] | None = (
            [Option(option) for option in data.get("options")]
            if data.get("options") is not None
            else None
        )
        self.guild_id: str | None = data.get("guild_id")
        self.target_id: str | None = data.get("target_id")

    def __iter__(self):
        if self.options is not None:
            yield from self.options
        else:
            yield None

    def __str__(self):
        return f"{self.name}"

    def __eq__(self, other):
        return self.id == other.id


class InteractionUtil:
    """Utility class intended for interactions"""

    def __init__(self, bot: Bot, http: http):
        self.bot: Bot = bot
        self.http: http = http

    async def interaction_search(
        self,
        query: str,
        channel_id: str,
        type: int = 1,
        cursor: str = None,
        bot_id: str = None,
        command_id: str = None,
    ) -> Search:
        """Search for interactions within a specific guild channel, you can specify certain parameters

        Args:
            query (str): Query to search for
            channel_id (str): Channel ID to search within
            type (int): Type of command to search for
            bot_id (str): Specify what bot specifically to search for
            command_id (str): Specify a command id to search for, to view options

        Returns:
            Search object
        """

        endpoint = (
            f"/channels/{channel_id}/application-commands/search?query={query}&limit=7"
        )
        endpoint += f"&type={type}"
        if cursor is not None:
            endpoint += f"&cursor={cursor}"
        if command_id is not None:
            endpoint += f"&command_id={command_id}"
        if bot_id is not None:
            endpoint += f"&application_id={bot_id}"
        data = await self.http.request("get", endpoint)
        return Search(
            data["application_commands"],
            data["cursor"]["previous"],
            data["cursor"]["next"],
        )

    async def trigger_slash(
        self,
        command: SlashCommand,
        channel_id: str,
        bot_id: str,
        value: list[str | None] | None = None,
        option: list[Option] | None = None,
        guild_id: str | None = None,
    ):
        nonce = int(time.time())
        payload = {
            "type": 2,
            "application_id": bot_id,
            "channel_id": channel_id,
            "nonce": nonce,
            "session_id": self.bot.session_id,
        }
        if guild_id is not None:
            payload["guild_id"] = guild_id
        data = {
            "version": command.version,
            "id": command.id,
            "name": command.name,
            "type": command.type,
            "options": [],
            "application_command": command.raw_data,
            "attachments": [],
        }
        if option is not None:
            dic = {"options": []}
            for index, (opt, value) in enumerate(zip_longest(option, value)):
                if index == 0:
                    if value is None:
                        dic["options"].append(
                            {"name": opt.name, "type": opt.type, "options": []}
                        )
                    else:
                        dic["options"].append(
                            {"name": opt.name, "type": opt.type, "value": value}
                        )
                elif value is not None:
                    for opten in dic["options"]:
                        opten["options"].append(
                            {"name": opt.name, "type": opt.type, "value": value}
                        )

            data |= dic
        payload["data"] = data
        randstr = "".join(random.sample(string.ascii_letters + string.digits, k=16))
        boundary_val = f"----WebkitFormBoundary{randstr}"
        req_data = f'--{boundary_val}\r\nContent-Disposition: form-data; name="payload_json"\r\n\r\n{json.dumps(payload)}\r\n--{boundary_val}--'
        await self.http.request(
            "post",
            "/interactions",
            headers={"content-type": f"multipart/form-data; boundary={boundary_val}"},
            data=req_data,
        )



from __future__ import annotations

import datetime
from base64 import b64encode
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..api.http import http
    from ..bot import Bot

class UserFlags:
    FLAGS = {
        "staff" : 1,
        "partner" : 2,
        "hypesquad" : 4,
        "bug_hunter" : 8,
        "mfa_sms" : 16,
        "premium_promo_dismissed" : 32,
        "hypesquad_bravery" : 64,
        "hypesquad_brilliance" : 128,
        "hypesquad_balance" : 256,
        "early_supporter" : 512,
        "team_user" : 1024,
        "system" : 4096,
        "has_unread_urgent_messages" : 8192,
        "bug_hunter_level_2" : 16384,
        "verified_bot" : 65536,
        "verified_bot_developer" : 131072,
        "discord_certified_moderator" : 262144,
        "bot_http_interactions" : 524288,
        "spammer" : 1048576,
        "disabled_nitro" : 2097152,
        "active_developer" : 4194304,
        "higher_ratelimits" : 8589934592,
        "deleted" : 17179869184,
        "deleted_sus_activity" : 34359738368,
        "self_deleted" : 68719476736,
        "premium_discriminator" : 137438953472,
        "used_desktop_client" : 274877906944,
        "used_web_client" : 549755813888,
        "used_mobile_client" : 1099511627776,
        "disabled" : 2199023255552,
        "verified" : 8796093022208,
        "quarantined" : 17592186044416,
        "collaborator_staff" : 1125899906842624,
        "restricted_collaborator" : 2251799813685248,
    }

    @classmethod
    def calculate_flags(cls, flag_value: int):
        return [
            key
            for key, value in cls.FLAGS.items()
            if (flag_value & value) == value
        ]

class Profile:
    def __init__(self, UserPayload: dict, bot: Bot, http: http) -> None:
        self.bot: Bot = bot
        self.http: http = http

        self.__update(UserPayload)

    def __update(self, data: dict):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """

        self.connected_accounts = [
            Connected_Account(account) for account in data.get("connected_accounts")
        ]

        self.mutual_guilds = [
            self.bot.get_guild(guild["id"]) for guild in data.get("mutual_guilds")
        ]

        self.id = data["user"]["id"]
        self.premium_type = data.get("premium_type")
        user_profile = data.get("user_profile")
        try:
            self.emoji = user_profile.get("emoji")
        except:
            self.emoji = None
        try:
            self.bio = user_profile.get("bio")
        except:
            self.bio = None
        try:
            self.accent_color = user_profile.get("accent_color")
        except:
            self.accent_color = None
        try:
            self.banner = user_profile.get("banner")
        except:
            self.banner = None
        self.banner_url = (
            f"https://cdn.discordapp.com/banners/{self.id}/{self.banner}.png?size=1024"
            if self.banner != None
            else None
        )


class Connected_Account:
    def __init__(self, data) -> None:
        self.__update(data)

    def __update(self, data: dict):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """
        self.type = data.get("type")
        self.name = data.get("name")
        self.id = data.get("id")


class User:
    """User Object"""

    def __init__(self, UserPayload: dict, bot: Bot, http: http) -> None:
        self.bot: Bot = bot

        self.http: http = http
        self._update(UserPayload)

    def __eq__(self, other):
        return self.id == other.id

    def __str__(self):
        return f"""{self.name}#{self.discriminator}"""

    @property
    def created_at(self) -> datetime.datetime:
        """Returns the time in which the User was created

        Returns:
            datetime.datetime: The timestamp
        """
        return datetime.datetime.utcfromtimestamp(
            ((int(self.id) >> 22) + 1420070400000) / 1000
        )

    @property
    def b64token(self) -> str:
        """Returns the b64 user id

        Returns:
            str: The b64 user id
        """
        return str(b64encode(self.id.encode("utf-8")), "utf-8")

    @property
    def flags(self) -> list[str]:
        """Returns the flags of the user as a string
        
        Returns:
            list[str]: The flags
        """
        return UserFlags.calculate_flags(self.raw_flags)

    @property
    def public_flags(self) -> list[str]:
        """Returns the public flags of the user as a string
        
        Returns:
            list[str]: The flags
        """
        return UserFlags.calculate_flags(self.raw_public_flags)

    def _update(self, data):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """
        self.name = data.get("username")
        self.id = data.get("id")
        self.discriminator = data.get("discriminator")
        self.avatar = data.get("avatar")
        self.banner = data.get("banner")
        self.accent_colour = data.get("accent_color")
        self.bot_acc = data.get("bot")
        self.avatar_url = (
            f"https://cdn.discordapp.com/avatars/{self.id}/{self.avatar}.png?size=4096"
            if self.avatar != None
            else None
        )
        self.banner_url = (
            f"https://cdn.discordapp.com/banners/{self.id}/{self.banner}.png?size=1024"
            if self.banner != None
            else None
        )
        self.system = data.get("system")
        self.raw_flags = data.get("flags") if data.get("flags") is not None else 0
        self.raw_public_flags = data.get("public_flags") if data.get("public_flags") is not None else 0

    async def create_dm(self):
        """Create a dm for the user"""
        await self.http.request(
            method="post",
            endpoint="/users/@me/channels",
            json={"recipients": [self.id]},
        )

    async def get_profile(self) -> Profile:
        """Get the User profile

        Returns:
            Profile: The User Profile object
        """
        data = await self.http.request(
            method="get", endpoint=f"/users/{self.id}/profile?with_mutual_guilds=true"
        )

        if data != None:
            data = Profile(data, self.bot, self.http)

        return data

    async def get_mutual_friends(self) -> list[User]:
        """Get the User mutual friends

        Returns:
            list[User]: Mutual friends
        """
        data = await self.http.request(
            method="get", endpoint=f"/users/{self.id}/relationships"
        )

        if len(data) != 0:
            return [User(user, self.bot, self.http) for user in data]
        else:
            return []



from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
if __name__ == "__main__":
    this_directory = Path(__file__).parent
    long_description = (this_directory / "README.md").read_text()
    setup(
        name="selfcord.py",
        packages=find_packages(
            include=[
                "selfcord",
                "selfcord.api",
                "selfcord.utils",
                "selfcord.models",
                "selfcord.api.voice",
            ]
        ),
        version="0.2.0",
        description="A Discord API wrapper designed for selfbots!",
        readme="README.md",
        author="Shell",
        extras_require={"voice": ["pynacl==1.5.0", "opuslib==3.0.1"]},
        license="MIT",
        install_requires=[
            "aiohttp==3.7.4.post0",
            "aioconsole==0.3.3",
            "websockets==10.1",
            "requests",
        ],
        setup_requires=["pytest-runner"],
        tests_require=["pytest"],
        test_suite="tests",
        keywords=["selfbot", "discord", "discordapi", "discordwrapper"],
        long_description=long_description,
        url="https://github.com/Shell1010/Selfcord",
        long_description_content_type="text/markdown",
    )



from __future__ import annotations

import asyncio
import contextlib
import importlib
import inspect
import io
import random
import time
from collections import defaultdict
from traceback import format_exception
from typing import TYPE_CHECKING

import aiohttp
from aioconsole import aexec

from .api import Activity, gateway, http
from .models import (Client, DMChannel, GroupChannel, Guild, InteractionUtil,
                     Option, Search, SlashCommand, TextChannel, User,
                     VoiceChannel)
from .utils import (Command, CommandCollection, Context, Event, Extension,
                    ExtensionCollection, logging)

if TYPE_CHECKING:
    from .api.gateway import gateway
    from .api.http import http

log = logging.getLogger("Bot")


class Bot:
    """Bot instance as entry point to interact with the bot

    Args:
        debug (bool): Whether to start the bot in debug mode, defaults to False.
        prefixes (list[str]): Prefixes for the bot, defaults to s!.
        inbuilt_help (bool): Whether the inbuilt help command should be enabled, defaults to True.
        userbot (bool): Whether the bot should be a userbot rather than selfbot, defaults to False.
        eval (bool): Whether to have the eval command as default, defaults to False.
    """

    def __init__(
        self,
        debug: bool = False,
        prefixes: list[str] = ["s!"],
        inbuilt_help: bool = True,
        userbot: bool = False,
        eval: bool = False,
    ) -> None:
        self.inbuilt_help: bool = inbuilt_help
        self.debug: bool = debug
        self.token = None
        self.http: http = http(debug)
        self.t1: float = time.perf_counter()
        self.session_id = None
        self.resume_url = None
        self.gateway: gateway = gateway(self.http, self.debug)
        self._events = defaultdict(list)
        self.commands = CommandCollection()
        self.prefixes: list[str] = (
            prefixes if isinstance(prefixes, list) else [prefixes]
        )
        self.extensions = ExtensionCollection()
        self.user = None
        self.eval: bool = eval
        self.userbot: bool = userbot

    def run(self, token: str):
        """Used to start connection to gateway as well as gather user information

        Args:
            token (str): _description_
        """
        self.token = token

        async def runner():
            data = await self.http.static_login(token)
            self.user = Client(data)
            await self.gateway.start(token, self.user, self)
            if self.debug:
                log.debug("Started Bot")
                log.info(f"Logged in as {self.user}")

        with contextlib.suppress(KeyboardInterrupt):
            asyncio.run(runner())

    @property
    def latency(self):
        """Latency of heartbeat ack, gateway latency essentially"""
        return self.gateway.latency

    # For events
    async def inbuilt_commands(self):
        """
        I call this on bot initialisation, it's the inbuilt help command
        """
        if self.inbuilt_help:

            @self.cmd("The help command!", aliases=["test"])
            async def help(ctx, cat=None):
                """The help command, dedicated to viewing all commands, extensions and information regarding commands."""
                if cat is None:
                    msg = f"```ini\n[ {self.user.name} Selfbot ]\n"
                    msg += f"[ {self.user} ]\nType <prefix>help <ext_name> to view commands relating to a specific extension. Type <prefix>help <cmd_name> to view information regarding a command.\n[ .Prefixes ] : {self.prefixes}\n\n"
                    msg += f"[ .Commands ]\n"
                    for command in self.commands:
                        msg += f". {command.name}: {command.description}\n"
                    msg += "\n[ .Extensions ]\n"
                    for ext in self.extensions:
                        msg += f"[ {ext.name} ] : [ {ext.description} ]\n"

                    msg += "```"
                    return await ctx.reply(f"{msg}")

                else:
                    name = cat.lower()
                    for ext in self.extensions:
                        if name == ext.name.lower():
                            msg = f"```ini\n[ {self.user.name} Selfbot ]\n"
                            msg += f"[ {self.user} ]\n\nType <prefix>help <ext_name> to view commands relating to a specific extension. Type <prefix>help <cmd_name> to view information regarding a command.\n\n[ .Prefixes ] : {self.prefixes}\n\n"
                            msg += f"[ .Commands ]\n"
                            for command in ext.commands:
                                if command.ext == ext.ext:
                                    msg += f". {command.name}: {command.description}\n"

                            msg += "```"
                            return await ctx.reply(f"{msg}")
                    else:
                        for cmd in self.commands:
                            if name == cmd.name.lower():
                                msg = f"```ini\n[ {self.user.name} Selfbot ]\n"
                                msg += f"[ {self.user} ]\n\nType <prefix>help <ext_name> to view commands relating to a specific extension. Type <prefix>help <cmd_name> to view information regarding a command.\n\n[ .Prefixes ] : {self.prefixes}\n\n"
                                msg += f"[ .{cmd.name} ]\n"
                                msg += f"[ Description ] :  {cmd.description} \n"
                                msg += f"[ Long Description ] :\n{cmd.func.__doc__}\n"
                                msg += f"[ Aliases ] : {cmd.aliases} \n"
                                args = inspect.signature(cmd.func)
                                msg += f"\n[ Example Usage ] :\n[ {self.prefixes[0]}{cmd.aliases[0]}"
                                for arg in args.parameters.keys():
                                    if arg in ["self", "ctx"]:
                                        continue
                                    msg += f" <{arg}>"
                                msg += " ]"

                                msg += "```"
                                return await ctx.reply(f"{msg}")
                        for ext in self.extensions:
                            for cmd in ext.commands:
                                if name == cmd.name.lower():
                                    msg = f"```ini\n[ {self.user.name} Selfbot ]\n"
                                    msg += f"[ {self.user} ]\n\nType <prefix>help <ext_name> to view commands relating to a specific extension. Type <prefix>help <cmd_name> to view information regarding a command.\n\n[ .Prefixes ] : {self.prefixes}\n\n"
                                    msg += f"[ .{cmd.name} ]\n"
                                    msg += f"[ Description ] :  {cmd.description} \n"
                                    msg += (
                                        f"[ Long Description ] :\n{cmd.func.__doc__}\n"
                                    )
                                    msg += f"[ Aliases ] :  {cmd.aliases} \n"
                                    args = inspect.signature(cmd.func)
                                    msg += f"\n[ Example Usage ] :\n[ {self.prefixes[0]}{cmd.aliases[0]}"
                                    for arg in args.parameters.keys():
                                        if arg in ["self", "ctx"]:
                                            continue
                                        msg += f" <{arg}>"
                                    msg += " ]"

                                    msg += "```"
                                    return await ctx.reply(f"{msg}")

        if self.eval:

            def clean_code(content):
                if content.startswith("```") and content.endswith("```"):
                    return "\n".join(content.split("\n")[1:])[:-3]
                else:
                    return content

            @self.cmd(description="Executes and runs code", aliases=["exec"])
            async def eval(ctx, *, code):
                """Runs python code via exec, intended for experienced usage. This can be DANGEROUS if you do not know what you are doing, use with caution."""
                code = clean_code(code)

                try:
                    with contextlib.redirect_stdout(io.StringIO()) as f:
                        await aexec(source=code, local=globals() )
                        result = f"```{f.getvalue()}\n```"
                except Exception as e:
                    error = "".join(format_exception(e, e, e.__traceback__))
                    result = f"```\n{error}```"

                await ctx.reply(result)

    def on(self, event: str):
        """Decorator for events

        Args:
            event (str): The event to check for
        """

        def decorator(coro):
            if not inspect.iscoroutinefunction(coro):
                log.error("Not a coroutine")
            else:
                self._events[event].append(Event(name=event, coro=coro, ext=None))

                def wrapper(*args, **kwargs):
                    result = self._events[event].append(
                        Event(name=event, coro=coro, ext=None)
                    )

                    return result

                return wrapper

        return decorator

    async def emit(self, event, *args, **kwargs):
        """Used to essentially push values to the decorator when the event fires

        Args:
            event (str): The event name
        """
        on_event = f"on_{event}"

        # try:
        if hasattr(self, on_event):
            await getattr(self, on_event)(*args, **kwargs)
        if event in self._events.keys():
            for Event in self._events[event]:
                if Event.coro.__code__.co_varnames[0] == "self":
                    asyncio.create_task(Event.coro(Event.ext, *args, **kwargs))

                else:
                    asyncio.create_task(Event.coro(*args, **kwargs))

    def cmd(self, description="", aliases=[]):
        """Decorator to add commands for the bot

        Args:
            description (str, optional): Description of command. Defaults to "".
            aliases (list, optional): Alternative names for command. Defaults to [].

        Raises:
            RuntimeWarning: If you suck and don't use a coroutine
        """
        if isinstance(aliases, str):
            aliases = [aliases]

        def decorator(coro):
            name = coro.__name__
            if not inspect.iscoroutinefunction(coro):
                log.error("Not a coroutine")
            else:
                cmd = Command(
                    name=name, description=description, aliases=aliases, func=coro
                )
                self.commands.add(cmd)
            return cmd

        return decorator

    def add_cmd(self, coro, description="", aliases=[]):
        """
        Function to add commands manually without decorator

        Args:
            coro (coroutine): The function to add
            description (str, optional): Description of command. Defaults to "".
            aliases (list, optional): Alternative names for command. Defaults to [].

        Raises:
            RuntimeWarning: If you suck and don't use a coroutine
        """
        if isinstance(aliases, str):
            aliases = [aliases]
        name = coro.__name__
        if not inspect.iscoroutinefunction(coro):
            log.error("Not a coroutine")
        else:
            cmd = Command(
                name=name, description=description, aliases=aliases, func=coro
            )
            self.commands.add(cmd)

    async def process_commands(self, msg):
        """
        What is called in order to actually get command input and run commands

        Args:
            msg (str): The message containing command
        """
        context = Context(self, msg, self.http)

        asyncio.create_task(context.invoke())

    async def load_extension(self, name: str):
        """
        Load various extensions/plugins/cogs if any.

        Args:
            name (str): Name of the extension to load

        Raises:
            ModuleNotFoundError: If you suck and don't know the name of what you want to load
        """
        try:
            name = importlib.util.resolve_name(name, None)
        except Exception as e:
            error = "".join(format_exception(e, e, e.__traceback__))
            log.error(f"Could not resolve extension name\n{error}")

        spec = importlib.util.find_spec(name)

        lib = importlib.util.module_from_spec(spec)

        try:
            spec.loader.exec_module(lib)
        except Exception as e:
            error = "".join(format_exception(e, e, e.__traceback__))
            log.error(f"Spec could not be loaded\n{error}")
        try:
            ext = getattr(lib, "Ext")
        except Exception as e:
            error = "".join(format_exception(e, e, e.__traceback__))
            log.error(f"Extension does not exist\n{error}")

        # Creates an Extension - ext in this case refers to the Ext class used for initialisation
        ext = Extension(
            name=ext.name,
            description=ext.description,
            ext=ext(self),
            _events=ext._events,
        )
        self.extensions.add(ext)
        try:
            for name, event in ext._events.items():
                for ext_event in event:
                    self._events[name].append(
                        Event(name=name, coro=ext_event.coro, ext=ext.ext)
                    )
            if self.debug:
                log.debug("Loaded Extension")
                log.info(f"Loaded Extension {ext.name} to Bot")

        except Exception as e:
            error = "".join(format_exception(e, e, e.__traceback__))
            log.error(f"Failed to load extension events\n{error}")

    async def trigger_slash(
        self,
        command: SlashCommand,
        channel_id: str,
        bot_id: str,
        value: list[str | None] | None = None,
        option: list[Option] | None = None,
        guild_id: str | None = None,
    ):
        interaction = InteractionUtil(self, self.http)
        await interaction.trigger_slash(
            command, channel_id, bot_id, value, option, guild_id
        )

    async def interaction_search(
        self,
        query: str,
        channel_id: str,
        type: int = 1,
        cursor: str = None,
        bot_id: str = None,
        command_id: str = None,
    ) -> Search:
        """Search for interactions within a specific guild channel, you can specify certain parameters

        Args:
            query (str): Query to search for
            channel_id (str): Channel ID to search within
            type (int): Type of command to search for
            bot_id (str): Specify what bot specifically to search for
            command_id (str): Specify a command id to search for, to view options

        Returns:
            Search object
        """
        interaction = InteractionUtil(self, self.http)
        return await interaction.interaction_search(
            query, channel_id, type, cursor, bot_id, command_id
        )

    def get_channel(self, channel_id: str):
        """
        Function to help retrieve channel from bot cache

        Args:
            channel_id (str): The channel id to search for

        Returns:
            Channel: The Channel object
        """
        for channel in self.user.private_channels:
            if channel_id == channel.id:
                return channel
        for guild in self.user.guilds:
            for channel in guild.channels:
                if channel_id == channel.id:
                    return channel

    def get_guild(self, guild_id: str):
        """
        Function to help retrieve guild from bot cache

        Args:
            guild_id (str): The guild id to search for

        Returns:
            Guild: The Guild object
        """
        for guild in self.user.guilds:
            if guild.id == guild_id:
                return guild

    async def get_user(self, user_id: str) -> User:
        """
        Function to retrieve user data. Probably need to be friends with them to retrieve the details.

        Args:
            user_id (Str): ID of the other user.

        Returns:

            User: The User object
        """

        data = await self.http.request(method="get", endpoint=f"/users/{user_id}")

        return User(data, bot=self, http=self.http)

    async def create_guild(
        self, name: str, icon_url: str = None, template: str = "2TffvPucqHkN"
    ):
        """Creates a guild"""
        image = None if icon_url is None else await self.http.encode_image(icon_url)
        json = await self.http.request(
            method="post",
            endpoint="/guilds",
            headers={
                "origin": "https://discord.com",
                "referer": "https://discord.com/channels/@me",
            },
            json={"name": name, "icon": image, "template": template},
        )
        if self.debug:
            log.debug("Finished Creating Guild")
            log.info(
                f"Created Guild NAME: {name} TEMPLATE: {template} ICON: {icon_url}"
            )
        return Guild(json, self, self.http)

    async def add_friend(self, user_id: str):
        """
        Function to add a specific user as a friend.

        Args:
            user_id (str): ID of the possible random user.

        Returns:
            No return value.
        """

        await self.http.request(
            method="put",
            endpoint=f"/users/@me/relationships/{user_id}",
            headers={
                "origin": "https://discord.com",
                "referer": f"https://discord.com/channels/@me/{random.choice(self.user.private_channels).id}",
            },
            json={},
        )
        if self.debug:
            log.debug("Sent Friend request")
            log.info(f"Sent Friend request to {user_id}")

    async def edit_profile(self, bio: str = None, accent: int = None):
        """Edits user profile"""
        fields = {}
        if bio != None:
            fields["bio"] = bio
        if accent != None:
            fields["accent"] = accent
        await self.http.request(
            method="patch", endpoint="/users/@me/profile", json=fields
        )
        if self.debug:
            log.debug("Finished Edit profile")

    async def change_pfp(self, avatar_url=None):
        """Disclaimer: This may phone lock your account :(

        Args:
            avatar_url (str): URL of image

        Raises:
            TypeError: URL not specified
        """
        if avatar_url != None:
            image = await self.http.encode_image(avatar_url)
            await self.http.request(
                method="patch",
                endpoint="/users/@me",
                headers={
                    "origin": "https://discord.com",
                    "referer": "https://discord.com/channels/@me",
                },
                json={"avatar": image},
            )
        else:
            log.error("Avatar URL not specified")
        if self.debug:
            log.debug("Finished changing avatar")

    async def create_dm(self, recipient_id: int):
        """
        Function to create new DM Channel with other user. Can be used with bots too.

        Args:
            recipient_id (snowflake): ID of recipient - Has to be user or bot ID

        Raises:
            TypeError: Recipient ID not specified

        Returns:
            DMChannel object
        """
        if recipient_id is not None:
            data = await self.http.request(
                method="post",
                endpoint="/users/@me/channels",
                json={"recipient_id": recipient_id},
            )
            if self.debug:
                log.debug("Created DM Channel")
            return DMChannel(data, bot=self, http=self.http)
        else:
            log.error("Recipient ID not specified")

    async def join_guild(self, code: str) -> Guild | None:
        """Helper function to join guilds
        
        Args:
            code (str): Invite Code
        Returns:
            Guild object | None
        """
        data = await self.http.request("post", f"/invites/{code}", json = {
            "session_id": f"{self.session_id}"
        }
        )
        if data is not None:
            return Guild(data['guild'], self, self.http)
        return None
    async def redeem_nitro(self, code: str):
        """Helper function to redeem nitro

        Args:
            code (str): Nitro code
        """
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"https://canary.discord.com/api/v9/entitlements/gift-codes/{code}/redeem",
                headers={
                    "authorization": f"{self.token}",
                    "user-agent": "Mozilla/5.0 (X11; Linux x86_64; rv:109.0) Gecko/20100101 Firefox/112.0",
                    "content-type": "application/json",
                },
                json={},
            ) as resp:
                j = await resp.json()
                log.info(f"Status: {resp.status} Nitro Redeem response: {j}")

    async def change_hypesquad(self, house: "str"):
        """Helper function to change hypesquad

        Args:
            house (str): Hypesquad name
        """
        if house.lower() == "bravery":
            await self.http.request(
                method="post", endpoint="/hypesquad/online", json={"house_id": 1}
            )
        if house.lower() == "brilliance":
            await self.http.request(
                method="post", endpoint="/hypesquad/online", json={"house_id": 2}
            )
        if house.lower() == "balance":
            await self.http.request(
                method="post", endpoint="/hypesquad/online", json={"house_id": 3}
            )
        if self.debug:
            log.debug("Finished changing hypesquad")
            log.info(f"Changed hypesquad to {house}")

    async def change_presence(self, status: str, afk: bool, activity: dict):
        """Change discord activity presence

        Args:
            status (str): Online, Offline, Dnd, Invisible
            afk (bool): True or False
            activity (dict): Selfcord.Activity method.
        """
        await self.gateway.change_presence(status, afk, activity=activity)
        if self.debug:
            log.debug("Finished changing presence")
            log.info(f"Changed Status to {status}, AFK to {afk}")



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .role import Role


class Member:
    """Member Object"""

    def __init__(self, UserPayload: dict) -> None:
        self.roles: list[Role] = []

        self._update(UserPayload)

    def __str__(self):
        return f"""{self.name}#{self.discriminator}"""

    def __eq__(self, other):
        return self.id == other.id

    def _update(self, data: dict):
        """Updater method intended to create the attributes for the object

        Args:
            data (dict): JSON data from gateway
        """
        user = data.get("user")
        self.name = user.get("username")
        self.id = user.get("id")
        self.discriminator = user.get("discriminator")
        self._avatar = user.get("avatar")
        self._banner = user.get("banner")
        self._accent_colour = user.get("accent_color")
        self._public_flags = user.get("public_flags")
        self.bot_acc = user.get("bot")
        self.joined_at = data.get("joined_at")
        self.nick = data.get("nick")
        self.system = data.get("system")



from __future__ import annotations

import inspect
import re
from collections import defaultdict
from traceback import format_exception
from typing import TYPE_CHECKING, Any

from .logging import logging

if TYPE_CHECKING:
    from ..api import *
    from ..bot import Bot
    from ..models import *


log = logging.getLogger("Commands")


class Extension:
    """Extension object. Discord.py equivalent of cogs, a helper system to help manage and organise code into multiple files"""

    def __init__(self, **kwargs):
        self.name: str | None = kwargs.get("name")
        self.description: str | None = kwargs.get("description")
        self.ext: Extension | None = kwargs.get("ext")
        self._events = defaultdict(list)
        _events: defaultdict = self.ext._events
        commands: CommandCollection = self.ext.commands
        self.commands = CommandCollection()
        for cmd in commands.recents():
            setattr(cmd, "ext", self.ext)
            self.commands.add(cmd)
        self.commands.copy()
        commands.clear()
        self._events.update(_events)
        _events.clear()


class ExtensionCollection:
    """Extension collection, where extensions are stored into. Utilised for Extender, Extensions as a whole. This is also used within help commands and command invocation."""

    def __init__(self):
        self.extensions = {}

    def __iter__(self):
        yield from self.extensions.values()

    def _is_already_registered(self, ext: Extension) -> bool:
        """Whether the specified Extension is already registered

        Args:
            ext (Extension): Extension to check

        Returns:
            bool: True or False
        """
        for extension in self.extensions.values():
            return ext.name == extension
        else:
            return False

    def add(self, ext: Extension):
        """Adds an extension

        Args:
            ext (Extension): Extension to add

        Raises:
            ValueError: Extension must be subclass of extension
            ValueError: A name or alias is already registered
        """
        if not isinstance(ext, Extension):
            log.error("ext is not an Extension or subclass of Extension")
        if self._is_already_registered(ext):
            log.error("Name or Alias is already registered")
        # Add extension to the collection
        self.extensions[ext.name] = ext

    def get(self, alias: str) -> Extension | None:
        """Get an extension

        Args:
            alias (str): Name of the extension

        Returns:
            Extension: Extension obtained
        """
        try:
            return self.extensions[alias]
        except KeyError as e:
            log.error(f"{e}")
        for extension in self.extensions:
            if extension.name in extension.aliases:
                return extension


class Command:
    """Command Object pretty much"""

    def __init__(self, **kwargs):
        self.name: str | None = kwargs.get("name")
        self.aliases: list[str] | None = [self.name] + kwargs.get("aliases", [])
        self.description: str | None = kwargs.get("description")
        self.func: function | None = kwargs.get("func")
        self.check: Any = inspect.signature(self.func).return_annotation
        self.signature = inspect.signature(self.func).parameters.items()


class CommandCollection:
    """Commands collection, where commands are stored into. Utilised for help commands and general command invocation."""

    def __init__(self, **kwargs):
        self.commands: dict[CommandCollection, function] = {}
        self.recent_commands: dict[CommandCollection, function] = {}

    def __len__(self):
        return len(self.commands)

    def __iter__(self):
        yield from self.commands.values()

    def _is_already_registered(self, cmd: Command) -> bool:
        """Whether the specified Command is already registered

        Args:
            cmd (Command): Command to check

        Returns:
            bool: True or False
        """
        for command in self.commands.values():
            for alias in cmd.aliases:
                return alias in command.aliases
            else:
                return False
        else:
            return False

    def append(self, collection):
        """Append to commands, and recent_commands

        Args:
            collection (CommandCollection): Collection instance

        Raises:
            ValueError: Collection must be subclass of CommandCollection
        """
        if not isinstance(collection, CommandCollection):
            log.error("collection is not a subclass of CommandCollection")
        for item in collection:
            self.commands[item.name] = item
            self.recent_commands[item.name] = item

    def add(self, cmd: Command):
        """Add a Command to the collection

        Args:
            cmd (Command): Command to be added

        Raises:
            ValueError: cmd must be a subclass of Command
            ValueError: Name or Alias is already registered
        """
        if not isinstance(cmd, Command):
            log.error("cmd must be a subclass of Command")
        if self._is_already_registered(cmd):
            log.error("Command Name or Alias is already registered")
        self.commands[cmd.name] = cmd
        self.recent_commands[cmd.name] = cmd

    def recents(self):
        """View commands recently acquired

        Yields:
            Generator: [Command]
        """
        yield from self.recent_commands.values()

    def copy(self):
        """Copy commands from recents to main collection"""
        self.commands.update(self.recent_commands)
        self.clear()

    def clear(self):
        """Clear recents"""
        self.recent_commands.clear()

    def get(self, alias) -> Command | None:
        """Get a specific command from the collection

        Args:
            alias (str): Name of the command

        Returns:
            Command: Command obtained
        """
        try:
            return self.commands[alias]
        except KeyError as e:
            log.error(f"{e}")
        for command in self.commands:
            if alias in command.aliases:
                return command


class Event:
    """Event object"""

    def __init__(self, name: str, coro, ext: Extension) -> None:
        self.name: str = name
        self.coro = coro
        self.ext: Extension = ext


class Extender:
    """Extender subclass for extensions, used for implementing the decorators."""

    commands = CommandCollection()
    _events: defaultdict = defaultdict(list)

    def __init_subclass__(cls, name: str, description: str = "") -> None:
        super().__init_subclass__()
        cls.name = name
        cls.description = description

    @classmethod
    def cmd(cls, description: str = "", aliases: list[str] = []):
        """Decorator to add commands for the bot

        Args:
            description (str, optional): Description of command. Defaults to "".
            aliases (list[str], optional): Alternative names for command. Defaults to [].

        Raises:
            RuntimeWarning: If you suck and don't use a coroutine
        """
        if isinstance(aliases, str):
            aliases = [aliases]

        def decorator(coro):
            name = coro.__name__
            if not inspect.iscoroutinefunction(coro):
                log.error("Not a coroutine")
            else:
                cmd = Command(
                    name=name, description=description, aliases=aliases, func=coro
                )
                cls.commands.add(cmd)

            return cmd

        return decorator

    @classmethod
    def on(cls, event: str):
        """Decorator for events

        Args:
            event (str): The event to check for
        """

        def decorator(coro):
            if not inspect.iscoroutinefunction(coro):
                log.error("Not a coroutine")
            else:
                eve = Event(name=event, coro=coro, ext=cls)
                cls._events[event].append(eve)

                def wrapper(*args, **kwargs):
                    result = cls._events[event].append(eve)
                    return result

                return wrapper

        return decorator

    @classmethod
    def add_cmd(cls, coro, description="", aliases=[]):
        """
        Function to add commands manually without decorator

        Args:
            coro (coroutine): The function to add
            description (str, optional): Description of command. Defaults to "".
            aliases (list, optional): Alternative names for command. Defaults to [].

        Raises:
            RuntimeWarning: If you suck and don't use a coroutine
        """
        if isinstance(aliases, str):
            aliases = [aliases]
        name = coro.__name__
        if not inspect.iscoroutinefunction(coro):
            log.error("Not a coroutine")
        else:
            cmd = Command(
                name=name, description=description, aliases=aliases, func=coro, ext=cls
            )
            cls.commands.add(cmd)


class Context:
    """Context related for commands, and invocation"""

    def __init__(self, bot, message, http) -> None:
        self.bot: Bot = bot
        self.message = message
        self.http = http

    @property
    def author(self) -> User:
        return self.message.author

    @property
    def guild(self) -> Guild:
        return self.message.guild

    @property
    def channel(self) -> Messageable | Voiceable:
        return self.message.channel

    @property
    def content(self) -> str:
        return self.message.content

    @property
    def command(self) -> function | None:
        if self.prefix is None:
            return None
        for command in self.bot.commands:
            for alias in command.aliases:
                if self.content.lower().startswith(self.prefix + alias):
                    return command
        for extension in self.bot.extensions:
            for command in extension.commands:
                for alias in command.aliases:
                    if self.content.lower().split(" ")[0] == self.prefix + alias:
                        self.extension = extension.ext
                        return command
        return None

    @property
    def alias(self) -> str | None:
        for command in self.bot.commands:
            for alias in command.aliases:
                if self.content.lower().startswith(self.prefix + alias.lower()):
                    return alias
        for extension in self.bot.extensions:
            for command in extension.commands:
                for alias in command.aliases:
                    if self.content.lower().startswith(self.prefix + alias.lower()):
                        self.extension = extension.ext
                        return alias
        return None

    @property
    def prefix(self) -> str | None:
        for prefix in self.bot.prefixes:
            if self.content.startswith(prefix):
                return prefix

    @property
    def command_content(self) -> str | None:
        """The content minus the prefix and command name, essentially the args

        Returns:
            str: String of content
        """
        if self.alias is None:
            return
        try:
            cut = len(self.prefix + self.alias)
            return self.content[cut:]
        except:
            return None

    def get_converter(self, param) -> type[str] | Any | None:
        if param.annotation is param.empty:
            return str
        if callable(param.annotation):
            return param.annotation
        else:
            log.error("Parameter annotation must be callable")

    def convert(self, param, value) -> str | Any:
        """Attempts to turn x value in y value, using get_converter func for the values

        Args:
            param (_type_): function parameter
            value (_type_): value in message

        Returns:
            Type[str]: The type of parameter
        """
        converter = self.get_converter(param)
        return converter(value)

    async def get_arguments(self) -> tuple[list, dict]:
        """Get arguments by checking function arguments and comparing to arguments in message.

        Returns:
            _type_: _description_
        """
        args: list[Any] = []
        kwargs: dict[Any, Any] = {}

        if self.command.signature is not None:
            signature = self.command.signature
        if self.command_content == "":
            return args, kwargs

        splitted = self.command_content.split(" ")[1:]

        for index, item in enumerate(splitted):
            user_regex = re.findall(r"<@[0-9]{18,19}>", item)
            if len(user_regex) > 0:
                x = re.findall(r"[0-9]{18,19}", item)
                if len(x) > 0:
                    val = x[0]
                    splitted[index] = val
            break
        for index, (name, param) in enumerate(signature):
            if name in ["ctx", "self"]:
                continue

            if param.kind is param.POSITIONAL_OR_KEYWORD:
                try:
                    arg: str | Any = self.convert(param, splitted.pop(0))
                    args.append(arg)
                except Exception:
                    pass
            if param.kind is param.VAR_KEYWORD:
                for arg in splitted:
                    arg = self.convert(param, arg)
                    args.append(arg)

            if param.kind is param.KEYWORD_ONLY:
                arg = self.convert(param, " ".join(splitted))
                kwargs[name] = arg

        for key in kwargs.copy():
            if not kwargs[key]:
                kwargs.pop(key)

        return args, kwargs

    async def invoke(self):
        """Used to actually run the command"""
        if self.command is None:
            return
        if not self.bot.userbot:
            if self.message.author.id != self.bot.user.id:
                return
        if self.command_content != None:
            args, kwargs = await self.get_arguments()
            func = self.command.func
            if func.__code__.co_varnames[0] == "self":
                args.insert(0, self.extension)
                args.insert(1, self)
            else:
                args.insert(0, self)
        try:
            await func(*args, **kwargs)
        except Exception as e:
            error = "".join(format_exception(e, e, e.__traceback__))
            log.error(f"Could not run command \n{error}")

    async def reply(self, content: str, file_paths: list[str] = [], tts=False) -> Message:
        """Helper function to reply to your own message containing the command

        Args:
            content (str): The message you would like to send
            tts (bool, optional): Whether message should be tts or not. Defaults to False.
        """
        message: Message = await self.channel.reply(self.message, content, file_paths, tts)
        return message

    async def send(self, content: str, file_paths: list[str] = [], tts=False) -> Message:
        """Helper function to send message to the current channel

        Args:
            content (str): The message you would like to send
            tts (bool, optional): Whether message should be tts or not. Defaults to False.
        """
        message: Message = await self.channel.send(content=content, file_paths=file_paths, tts=tts)
        return message

    async def spam(self, amount: int, content: str, file_paths: list[str] = [], tts: bool = False):
        """Helper function to spam messages in the current channel (uses asyncio.gather !!!!)

        Args:
            amount (int): Amount of messages to spam
            content (str): The message you would like to send
        """
        await self.channel.spam(amount, content, file_paths, tts)

    async def purge(self, amount: int = 0):
        """Helper function to purge messages in the current channel, uses asyncio gather.

        Args:
            amount (int): The amount of messages to purge, defaults to All.
        """
        await self.channel.purge(amount)

    async def edit(self, content: str, file_paths: list[str] = []) -> Message:
        """Helper function to edit the message you sent

        Args:
            content (str): Content to edit to
        """
        message = await self.message.edit(content, file_paths)
        return message



import logging
import logging.config

logging.config.dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": True,
    }
)

logging.getLogger("websockets.client").disabled = True
logging.getLogger("urllib3.connectionpool").disabled = True

FMT = "[{asctime}][{levelname:^9}] {name}: {message}"

FORMATS = {
    logging.DEBUG: f"\33[93m{FMT}\33[0m",
    logging.INFO: f"\33[34m{FMT}\33[0m",
    logging.WARNING: f"\33[33m{FMT}\33[0m",
    logging.ERROR: f"\33[31m{FMT}\33[0m",
    logging.CRITICAL: "\33[31m{message}\33[0m",
}


class CustomFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        log_fmt = FORMATS[record.levelno]
        formatter = logging.Formatter(log_fmt, style="{", datefmt="%H:%M:%S")
        return formatter.format(record)


handler = logging.StreamHandler()
handler.setFormatter(CustomFormatter())

logging.basicConfig(
    level=logging.DEBUG,
    handlers=[handler],
)
